# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from .core import IonType
from .simple_types import IonPyList, IonPyDict, IonPyNull, IonPyBool, IonPyInt, IonPyFloat, IonPyDecimal, \
    IonPyTimestamp, IonPyText, IonPyBytes, IonPySymbol
from base64 import standard_b64encode
import sys
if hasattr(sys, "pypy_version_info"):
    raise ImportError("The json_encoder module is not supported for use with PyPy.")
from jsonconversion.encoder import JSONExtendedEncoder


class IonToJSONEncoder(JSONExtendedEncoder):
    """JSON Encoder for Ion value types. Used in the json.dumps method as the cls parameter to support JSON encoding of
    Python Ion types: json.dumps(obj, cls=IonToJSONEncoder)

    Notes:
        The json_encoder module is not supported for use with PyPy.
    """

    def isinstance(self, obj, cls):
        if (not hasattr(sys, "pypy_version_info")) and \
                isinstance(obj, (IonPyList, IonPyDict, IonPyNull, IonPyBool, IonPyInt, IonPyFloat, IonPyDecimal,
                                 IonPyTimestamp, IonPyText, IonPyBytes, IonPySymbol)):
            return False
        return isinstance(obj, cls)

    def default(self, o):
        if (isinstance(o, IonPyInt) or isinstance(o, IonPyBool)) and o.ion_type == IonType.BOOL:
            return True if o == 1 else False
        elif isinstance(o, IonPyInt) and o.ion_type == IonType.INT:
            return int(o)
        elif isinstance(o, IonPyList) and (o.ion_type == IonType.LIST or o.ion_type == IonType.SEXP):
            return list(map(self.default, o))
        elif isinstance(o, IonPyDict) and o.ion_type == IonType.STRUCT:
            return {key: self.default(o[key]) for key in o.keys()}
        elif isinstance(o, IonPyNull):
            return None
        elif isinstance(o, IonPyBytes) and o.ion_type == IonType.BLOB:
            return standard_b64encode(o).decode("utf-8")
        elif isinstance(o, IonPyBytes) and o.ion_type == IonType.CLOB:
            return "".join(chr(b) for b in o)
        elif isinstance(o, IonPyDecimal) and o.ion_type == IonType.DECIMAL:
            return float(o)
        elif isinstance(o, IonPySymbol) and o.ion_type == IonType.SYMBOL:
            return o.text
        elif isinstance(o, IonPyTimestamp) and o.ion_type == IonType.TIMESTAMP:
            return str(o)
        elif isinstance(o, IonPyText) and o.ion_type == IonType.STRING:
            return str(o)
        elif isinstance(o, IonPyFloat) and o.ion_type == IonType.FLOAT:
            if "inf" in str(o) or "nan" in str(o):
                return None
            return float(o)
        else:
            super(IonToJSONEncoder, self).default(o)



#!/usr/bin/env python3
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
A collection of tools to help with protocol buffers for use in the Ion Benchmark CLI.

Usage:
    proto_tools.py <command> [<args>...]
    proto_tools.py --help

Commands:{commands}

Options:
     -h, --help
"""

import types
from docopt import docopt as _docopt
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from self_describing_proto import SelfDescribingProtoSerde


def wrap_command():
    """
    Wrap a protocol buffer in a SelfDescribingMessage protocol buffer.

    Usage:
        proto_tools.py wrap <schema_descriptor_file> <type_name> <input_file> [<output_file>]

    Arguments:
        <schema_descriptor_file>    A Protobuf FileDescriptorSet, generated using `protoc --descriptor_set_out ...`.
        <type_name>                 The type of <input_file>. Must be included in <schema_descriptor_file>.
        <input_file>                A protocol buffer to be wrapped in a self-describing protocol buffer.
        <output_file>               Where the wrapped protocol buffer should be saved. Default: `<input_file>.wrapped`
    """
    arguments = _docopt(wrap_command.__doc__, help=True)
    schema_file = arguments['<schema_descriptor_file>']
    type_name = arguments['<type_name>']
    input_file = arguments['<input_file>']
    output_file = arguments['<output_file>'] or "{}.wrapped".format(input_file)

    descriptor_set = _descriptor_pb2.FileDescriptorSet()

    with open(schema_file, "rb") as schema_fp:
        descriptor_set.ParseFromString(schema_fp.read())

    sd_proto = SelfDescribingProtoSerde()
    inner_obj = sd_proto.generate_class_definition(type_name, descriptor_set)()

    with open(input_file, "rb") as fp:
        inner_obj.ParseFromString(fp.read())
    with open(output_file, "wb") as fp:
        sd_proto.dump(inner_obj, fp)


def unwrap_command():
    """
    Unwrap the inner protocol buffer of a SelfDescribingMessage protocol buffer.

    Usage:
        proto_tools.py unwrap <input_file> [<output_file>]

    Arguments:
        <input_file>     The file to unwrap
        <output_file>    The destination for the unwrapped file. Default: `<input_file>.unwrapped`
    """
    arguments = _docopt(wrap_command.__doc__, help=True)
    input_file = arguments['<input_file>']
    output_file = arguments['<output_file>'] or "{}.unwrapped".format(input_file)

    sd_proto = SelfDescribingProtoSerde()
    with open(input_file, "rb") as fp:
        obj = sd_proto.load(fp)
    with open(output_file, "wb") as fp:
        fp.write(obj.SerializeToString())


def _list_commands():
    """
    Get a formatted list of commands as a string.
    """
    commands = ""
    for name, member in [*globals().items()]:
        if isinstance(member, types.FunctionType) and name.endswith("_command"):
            display_name = name.removesuffix("_command")
            # Get the first line of __doc__ to use as the summary.
            description = next((s.strip() for s in (member.__doc__ or "").split('\n') if s.strip()), "")
            commands += "\n    {: <16}{}".format(display_name, description)
    return commands


if __name__ == '__main__':
    docs = __doc__.format(commands=_list_commands())
    args = _docopt(docs, help=True, options_first=True)
    func_name = "{}_command".format(args['<command>'])
    if func_name in globals().items().mapping:
        globals().items().mapping[func_name]()
    else:
        exit("%r is not a proto_tools.py command. See 'proto_tools.py --help'." % args['<command>'])



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

import sys
import textwrap
from decimal import Decimal
from io import BytesIO, StringIO

import pytest

from amazon.ion import simpleion
from amazon.ion.core import IonEventType, IonType, IonEvent, ION_STREAM_END_EVENT
from amazon.ion.reader import read_data_event, NEXT_EVENT, blocking_reader, SKIP_EVENT
from amazon.ion.reader_binary import binary_reader
from amazon.ion.reader_managed import managed_reader
from amazon.ion.reader_text import text_reader
from amazon.ion.simple_types import IonPyFloat
from amazon.ion.symbols import shared_symbol_table, SymbolTableCatalog
from amazon.ion.writer import WriteEventType, blocking_writer
from amazon.ion.writer_binary import binary_writer
from amazon.ion.writer_text import text_writer
from amazon.ion.simpleion import c_ext

# Tests for the Python examples in the cookbook (https://amazon-ion.github.io/ion-docs/guides/cookbook.html).
# Changes to these tests should only be made in conjunction with changes to the cookbook examples.

if sys.version_info < (3, 6):
    pytest.skip(u'To avoid polluting the examples with extra compatibility code, they are only written for and tested '
                u'with Python 3.6+.')


def test_reading_simpleion_loads():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = u'{hello: "world"}'
    value = simpleion.loads(data)
    assert u'hello world' == u'hello %s' % value[u'hello']


def test_reading_simpleion_load():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = BytesIO(b'{hello: "world"}')
    value = simpleion.load(data)
    assert u'hello world' == u'hello %s' % value[u'hello']


def test_writing_simpleion_dumps():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = u'{hello: "world"}'
    value = simpleion.loads(data)
    ion = simpleion.dumps(value, binary=False)
    assert u'$ion_1_0 {hello:"world"}' == ion


def test_writing_simpleion_dump():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = u'{hello: "world"}'
    value = simpleion.loads(data)
    ion = BytesIO()
    simpleion.dump(value, ion, binary=True)
    assert b'\xe0\x01\x00\xea\xec\x81\x83\xde\x88\x87\xb6\x85hello\xde\x87\x8a\x85world' == ion.getvalue() \
           or b'\xe0\x01\x00\xea\xeb\x81\x83\xd8\x87\xb6\x85hello\xd7\x8a\x85world' == ion.getvalue()


def test_reading_simpleion_loads_multiple_top_level_values():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = u'1 2 3'
    value = simpleion.loads(data, single_value=False)
    assert [1, 2, 3] == value


def test_writing_simpleion_dumps_multiple_top_level_values():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = u'1 2 3'
    value = simpleion.loads(data, single_value=False)
    ion = simpleion.dumps(value, sequence_as_stream=True, binary=False)
    assert u'$ion_1_0 1 2 3' == ion


def test_reading_events_non_blocking():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    reader = managed_reader(text_reader())
    event = reader.send(NEXT_EVENT)
    # No data has been provided, so the reader is at STREAM_END
    # and will wait for data.
    assert event.event_type == IonEventType.STREAM_END
    # Send an incomplete Ion value.
    event = reader.send(read_data_event(b'{hello:'))
    # Enough data was available for the reader to determine that
    # the start of a struct value has been encountered.
    assert event.event_type == IonEventType.CONTAINER_START
    assert event.ion_type == IonType.STRUCT
    # Advancing the reader causes it to step into the struct.
    event = reader.send(NEXT_EVENT)
    # The reader reached the end of the data before completing
    # a value. Therefore, an INCOMPLETE event is returned.
    assert event.event_type == IonEventType.INCOMPLETE
    # Send the rest of the value.
    event = reader.send(read_data_event(b'"world"}'))
    # The reader now finishes parsing the value within the struct.
    assert event.event_type == IonEventType.SCALAR
    assert event.ion_type == IonType.STRING
    hello = event.field_name.text
    world = event.value
    # Advance the reader past the string value.
    event = reader.send(NEXT_EVENT)
    # The reader has reached the end of the struct.
    assert event.event_type == IonEventType.CONTAINER_END
    # Advancing the reader causes it to step out of the struct.
    event = reader.send(NEXT_EVENT)
    # There is no more data and a value has been completed.
    # Therefore, the reader conveys STREAM_END.
    assert event.event_type == IonEventType.STREAM_END
    assert u'hello world' == u'%s %s' % (hello, world)


def test_writing_events_non_blocking():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    def drain_data(incremental_event):
        incremental_data = b''
        while incremental_event.type == WriteEventType.HAS_PENDING:
            incremental_data += incremental_event.data
            incremental_event = writer.send(None)
        return incremental_data

    writer = binary_writer()
    event = writer.send(IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT))
    data = drain_data(event)
    event = writer.send(IonEvent(IonEventType.SCALAR, IonType.STRING, field_name=u'hello', value=u'world'))
    data += drain_data(event)
    event = writer.send(IonEvent(IonEventType.CONTAINER_END))
    data += drain_data(event)
    event = writer.send(ION_STREAM_END_EVENT)
    data += drain_data(event)
    assert b'\xe0\x01\x00\xea\xec\x81\x83\xde\x88\x87\xb6\x85hello\xde\x87\x8a\x85world' == data


def test_reading_events_blocking():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = BytesIO(b'{hello: "world"}')
    reader = blocking_reader(managed_reader(text_reader()), data)
    event = reader.send(NEXT_EVENT)
    assert event.event_type == IonEventType.CONTAINER_START
    assert event.ion_type == IonType.STRUCT
    # Advancing the reader causes it to step into the struct.
    event = reader.send(NEXT_EVENT)
    assert event.event_type == IonEventType.SCALAR
    assert event.ion_type == IonType.STRING
    hello = event.field_name.text
    world = event.value
    # Advance the reader past the string value.
    event = reader.send(NEXT_EVENT)
    # The reader has reached the end of the struct.
    assert event.event_type == IonEventType.CONTAINER_END
    # Advancing the reader causes it to step out of the struct.
    event = reader.send(NEXT_EVENT)
    # There is no more data and a value has been completed.
    # Therefore, the reader conveys STREAM_END.
    assert event.event_type == IonEventType.STREAM_END
    assert u'hello world' == u'%s %s' % (hello, world)


def test_writing_events_blocking():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-and-writing-ion-data
    data = BytesIO()
    writer = blocking_writer(binary_writer(), data)
    event_type = writer.send(IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT))
    # The value is not complete, so more events are required.
    assert event_type == WriteEventType.NEEDS_INPUT
    event_type = writer.send(IonEvent(IonEventType.SCALAR, IonType.STRING, field_name=u'hello', value=u'world'))
    # The value is not complete, so more events are required.
    assert event_type == WriteEventType.NEEDS_INPUT
    event_type = writer.send(IonEvent(IonEventType.CONTAINER_END))
    # The value is not complete, so more events are required.
    assert event_type == WriteEventType.NEEDS_INPUT
    event_type = writer.send(ION_STREAM_END_EVENT)
    # The end of the stream was signaled, so the data has been flushed.
    assert event_type == WriteEventType.COMPLETE
    assert b'\xe0\x01\x00\xea\xec\x81\x83\xde\x88\x87\xb6\x85hello\xde\x87\x8a\x85world' == data.getvalue()


def test_pretty_print_simpleion():
    if c_ext:
        return
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#pretty-printing
    unformatted = u'{level1: {level2: {level3: "foo"}, x: 2}, y: [a,b,c]}'
    value = simpleion.loads(unformatted)
    pretty = simpleion.dumps(value, binary=False, indent=u'  ')
    assert pretty == textwrap.dedent(u'''    $ion_1_0
    {
      level1: {
        level2: {
          level3: "foo"
        },
        x: 2
      },
      y: [
        a,
        b,
        c
      ]
    }''')


def test_pretty_print_events():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#pretty-printing
    pretty = BytesIO()
    writer = blocking_writer(text_writer(indent=u'  '), pretty)
    writer.send(ION_STREAM_END_EVENT)


def test_read_numerics_simpleion():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-numeric-types
    data = u'1.23456 1.2345e6 123456 12345678901234567890'
    values = simpleion.loads(data, single_value=False)
    assert isinstance(values[0], Decimal)
    assert isinstance(values[1], float)
    assert isinstance(values[2], int)
    assert isinstance(values[3], int)


def test_write_numeric_with_annotation_simpleion():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-numeric-types
    value = IonPyFloat.from_value(IonType.FLOAT, 123, (u'abc',))
    data = simpleion.dumps(value, binary=False)
    assert data == u'$ion_1_0 abc::123e+0' or data == u'$ion_1_0 abc::123.0e0'


def test_read_numerics_events():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-numeric-types
    data = BytesIO(b'1.23456 1.2345e6 123456 12345678901234567890')
    reader = blocking_reader(managed_reader(text_reader()), data)
    event = reader.send(NEXT_EVENT)
    assert event.ion_type == IonType.DECIMAL
    assert isinstance(event.value, Decimal)
    event = reader.send(NEXT_EVENT)
    assert event.ion_type == IonType.FLOAT
    assert isinstance(event.value, float)
    event = reader.send(NEXT_EVENT)
    assert event.ion_type == IonType.INT
    assert isinstance(event.value, int)
    event = reader.send(NEXT_EVENT)
    assert event.ion_type == IonType.INT
    assert isinstance(event.value, int)


def test_write_numeric_with_annotation_events():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#reading-numeric-types
    event = IonEvent(IonEventType.SCALAR, IonType.FLOAT, annotations=(u'abc',), value=123.0)
    data = BytesIO()
    writer = blocking_writer(text_writer(), data)
    writer.send(event)
    writer.send(ION_STREAM_END_EVENT)
    assert u'abc::123.0e0' == data.getvalue().decode(u'utf-8')


def sparse_reads_data():
    data = u'''
         $ion_1_0
         foo::{
           quantity: 1
         }
         bar::{
           name: "x",
           id: 7
         }
         baz::{
           items:["thing1", "thing2"]
         }
         foo::{
           quantity: 19
         }
         bar::{
           name: "y",
           id: 8
         }'''
    data = simpleion.dumps(simpleion.loads(data, single_value=False), sequence_as_stream=True)
    # This byte literal is included in the examples.
    assert data == b'\xe0\x01\x00\xea' \
                   b'\xee\xa5\x81\x83\xde\xa1\x87\xbe\x9e\x83foo\x88quantity\x83' \
                   b'bar\x82id\x83baz\x85items\xe7\x81\x8a\xde\x83\x8b!\x01\xea' \
                   b'\x81\x8c\xde\x86\x84\x81x\x8d!\x07\xee\x95\x81\x8e\xde\x91' \
                   b'\x8f\xbe\x8e\x86thing1\x86thing2\xe7\x81\x8a\xde\x83\x8b!' \
                   b'\x13\xea\x81\x8c\xde\x86\x84\x81y\x8d!\x08' \
           or data == b'\xe0\x01\x00\xea' \
                      b'\xee\xa5\x81\x83\xde\xa1\x87\xbe\x9e\x83foo\x88quantity\x83' \
                      b'bar\x82id\x83baz\x85items\xe6\x81\x8a\xd3\x8b!\x01\xe9' \
                      b'\x81\x8c\xd6\x84\x81x\x8d!\x07\xee\x95\x81\x8e\xde\x91' \
                      b'\x8f\xbe\x8e\x86thing1\x86thing2\xe6\x81\x8a\xd3\x8b!' \
                      b'\x13\xe9\x81\x8c\xd6\x84\x81y\x8d!\x08'
    return data


def test_sparse_reads_simpleion():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#performing-sparse-reads
    data = sparse_reads_data()  # The binary Ion equivalent of the above data.
    values = simpleion.loads(data, single_value=False)
    sum = 0
    for value in values:
        if u'foo' == value.ion_annotations[0].text:
            sum += value[u'quantity']
    assert 20 == sum


def test_sparse_reads_events():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#performing-sparse-reads
    data = BytesIO(sparse_reads_data())
    reader = blocking_reader(managed_reader(binary_reader()), data)
    sum = 0
    event = reader.send(NEXT_EVENT)
    while event != ION_STREAM_END_EVENT:
        assert event.event_type == IonEventType.CONTAINER_START
        assert event.ion_type == IonType.STRUCT
        if u'foo' == event.annotations[0].text:
            # Step into the struct.
            event = reader.send(NEXT_EVENT)
            while event.event_type != IonEventType.CONTAINER_END:
                if u'quantity' == event.field_name.text:
                    sum += event.value
                event = reader.send(NEXT_EVENT)
            # Step out of the struct.
            event = reader.send(NEXT_EVENT)
        else:
            # Skip over the struct without parsing its values.
            event = reader.send(SKIP_EVENT)
            assert event.event_type == IonEventType.CONTAINER_END
            # Position the reader at the start of the next value.
            event = reader.send(NEXT_EVENT)
    assert 20 == sum


def get_csv_structs():
    data = StringIO(
        u'''id,type,state
        1,foo,false
        2,bar,true
        3,baz,true'''
    )
    lines = data.readlines()[1:]

    def split_line(line):
        tokens = line.split(u',')
        mapping = (
            (u'id', int(tokens[0])),
            (u'type', tokens[1]),
            (u'state', u'true' == tokens[2].strip())
        )
        return dict(mapping)

    return [split_line(line) for line in lines]


def test_convert_csv_simpleion():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#converting-non-hierarchical-data-to-ion
    structs = get_csv_structs()
    ion = simpleion.dumps(structs, sequence_as_stream=True)
    assert ion == b'\xe0\x01\x00\xea\xee\x95\x81\x83\xde\x91\x87\xbe\x8e\x82id\x84type\x85state\xde\x8a\x8a!' \
                  b'\x01\x8b\x83foo\x8c\x10\xde\x8a\x8a!\x02\x8b\x83bar\x8c\x11\xde\x8a\x8a!\x03\x8b\x83baz\x8c\x11' \
           or \
           ion == b'\xe0\x01\x00\xea\xee\x95\x81\x83\xde\x91\x87\xbe\x8e\x82id\x84type\x85state\xda\x8a!' \
                  b'\x01\x8b\x83foo\x8c\x10\xda\x8a!\x02\x8b\x83bar\x8c\x11\xda\x8a!\x03\x8b\x83baz\x8c\x11'


def test_convert_csv_events():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#converting-non-hierarchical-data-to-ion
    structs = get_csv_structs()
    ion = BytesIO()
    writer = blocking_writer(binary_writer(), ion)
    for struct in structs:
        writer.send(IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT))
        writer.send(IonEvent(IonEventType.SCALAR, IonType.INT, field_name=u'id', value=struct[u'id']))
        writer.send(IonEvent(IonEventType.SCALAR, IonType.STRING, field_name=u'type', value=struct[u'type']))
        writer.send(IonEvent(IonEventType.SCALAR, IonType.BOOL, field_name=u'state', value=struct[u'state']))
        writer.send(IonEvent(IonEventType.CONTAINER_END))
    writer.send(ION_STREAM_END_EVENT)
    assert b'\xe0\x01\x00\xea\xee\x95\x81\x83\xde\x91\x87\xbe\x8e\x82id\x84type\x85state\xde\x8a\x8a!' \
           b'\x01\x8b\x83foo\x8c\x10\xde\x8a\x8a!\x02\x8b\x83bar\x8c\x11\xde\x8a\x8a!\x03\x8b\x83baz\x8c\x11' \
           == ion.getvalue()


def write_with_shared_symbol_table_simpleion():
    structs = get_csv_structs()
    table = shared_symbol_table(u'test.csv.columns', 1, (u'id', u'type', u'state'))
    data = simpleion.dumps(structs, imports=(table,), sequence_as_stream=True)
    # This byte literal is included in the examples.
    assert data == b'\xe0\x01\x00\xea' \
                   b'\xee\xa4\x81\x83\xde\xa0\x86\xbe\x9b\xde\x99\x84\x8e\x90' \
                   b'test.csv.columns\x85!\x01\x88!\x03\x87\xb0\xde\x8a\x8a!' \
                   b'\x01\x8b\x83foo\x8c\x10\xde\x8a\x8a!\x02\x8b\x83bar\x8c' \
                   b'\x11\xde\x8a\x8a!\x03\x8b\x83baz\x8c\x11'
    return data


def test_write_with_shared_symbol_table_simpleion():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#using-a-shared-symbol-table
    write_with_shared_symbol_table_simpleion()


def write_with_shared_symbol_table_events():
    structs = get_csv_structs()
    table = shared_symbol_table(u'test.csv.columns', 1, (u'id', u'type', u'state'))
    data = BytesIO()
    writer = blocking_writer(binary_writer(imports=(table,)), data)
    for struct in structs:
        writer.send(IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT))
        writer.send(IonEvent(IonEventType.SCALAR, IonType.INT, field_name=u'id', value=struct[u'id']))
        writer.send(IonEvent(IonEventType.SCALAR, IonType.STRING, field_name=u'type', value=struct[u'type']))
        writer.send(IonEvent(IonEventType.SCALAR, IonType.BOOL, field_name=u'state', value=struct[u'state']))
        writer.send(IonEvent(IonEventType.CONTAINER_END))
    writer.send(ION_STREAM_END_EVENT)
    return data.getvalue()


def test_write_with_shared_symbol_table_events():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#using-a-shared-symbol-table
    write_with_shared_symbol_table_events()


def test_read_with_shared_symbol_table_simpleion():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#using-a-shared-symbol-table
    data = write_with_shared_symbol_table_simpleion()
    table = shared_symbol_table(u'test.csv.columns', 1, (u'id', u'type', u'state'))
    catalog = SymbolTableCatalog()
    catalog.register(table)
    values = simpleion.loads(data, catalog=catalog, single_value=False)
    assert values[2][u'id'] == 3


def test_read_with_shared_symbol_table_events():
    # https://amazon-ion.github.io/ion-docs/guides/cookbook.html#using-a-shared-symbol-table
    table = shared_symbol_table(u'test.csv.columns', 1, (u'id', u'type', u'state'))
    catalog = SymbolTableCatalog()
    catalog.register(table)
    data = BytesIO(write_with_shared_symbol_table_simpleion())
    reader = blocking_reader(managed_reader(binary_reader(), catalog=catalog), data)
    # Position the reader at the first struct.
    reader.send(NEXT_EVENT)
    # Skip over the struct.
    reader.send(SKIP_EVENT)
    # Position the reader at the second struct.
    reader.send(NEXT_EVENT)
    # Skip over the struct.
    reader.send(SKIP_EVENT)
    # Position the reader at the third struct.
    event = reader.send(NEXT_EVENT)
    assert event.ion_type == IonType.STRUCT
    # Step into the struct
    event = reader.send(NEXT_EVENT)
    assert u'id' == event.field_name.text
    assert 3 == event.value



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Common test utilities for the ``tests`` package."""

import pytest

from contextlib import contextmanager


def listify(iter_func):
    """Takes an iterator function and returns a function that materializes it into a list."""
    def delegate(*args, **kwargs):
        return list(iter_func(*args, **kwargs))
    delegate.__name__ = iter_func.__name__
    return delegate


@contextmanager
def noop_manager():
    """A no-op context manager"""
    yield


def is_exception(val):
    """Returns whether a given value is an exception type (not an instance)."""
    return isinstance(val, type) and issubclass(val, Exception)


def parametrize(*values):
    """Idiomatic test parametrization.

    Parametrizes single parameter testing functions.
    The assumption is that the function decorated uses something
    like a ``namedtuple`` as its sole argument.

    Makes the ``id`` string be based on the parameter value's ``__str__``
    method.

    Examples:
        Usage of this decorator typically looks something like::

            class _P(namedtuple('Params', 'a, b, c')):
                def __str__(self):
                    return '{p.a} - {p.b} - {p.c}'.format(p=self)

            @tests.parametrize(
                _P(a='something', b=6, c=True),
                _P(s='some-other-thing', b=7, c=False),
                _P('look ma, positional', 9, True)
            def my_test(p):
                assert isinstance(p.a, str)
                assert p.b > 0
                assert isinstance(p.c, bool)

        :func:`amazon.ion.util.record` is also appropriate for the parameter
        class/tuple::

            class _P(amazon.ion.util.record('a', 'b', 'c')):
                def __str__(self):
                    return '{p.a} - {p.b} - {p.c}'.format(p=self)

    Args:
        values (Sequence[Any]): A sequence of values to pass to a single argument
            function.

    Returns:
        pytest.mark.parametrize: The decorator.
    """
    values = tuple((value,) for value in values)
    def decorator(func):
        if func.__code__.co_argcount != 1:
            raise ValueError('Expected a function with a single parameter.')
        argname = func.__code__.co_varnames[0]
        real_decorator = pytest.mark.parametrize(
            argnames=[argname],
            argvalues=values,
            ids=lambda x: str(x).replace('.', '_')
        )
        return real_decorator(func)

    return decorator



import os
import time
from os.path import abspath, join, dirname

from amazon.ion import simpleion
from amazon.ion.equivalence import ion_equals
from amazon.ionbenchmark import Format
from amazon.ionbenchmark.Format import format_is_ion, format_is_cbor, format_is_json, rewrite_file_to_format
from amazon.ionbenchmark.ion_benchmark_cli import TOOL_VERSION
from tests import parametrize


def generate_test_path(p):
    return join(dirname(abspath(__file__)), 'benchmark_sample_data', p)


def run_cli(c):
    import subprocess
    cmd = ["python", "./amazon/ionbenchmark/ion_benchmark_cli.py"] + c
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, text=True)
    error_code = proc.wait()
    (out, err) = proc.communicate()
    return error_code, out, err


def test_option_version():
    (error_code, out, _) = run_cli(["--version"])
    assert not error_code
    assert out.strip() == TOOL_VERSION


@parametrize(
    ('write', 'buffer'),
    ('write', 'file'),
    ('read', 'buffer'),
    ('read', 'file'),
)
def test_run_benchmark_spec(args):
    (command, io_type) = args
    (error_code, out, _) = run_cli(['spec', './tests/benchmark_sample_data/sample_spec/spec.ion', '-O', f'{{io_type:{io_type},command:{command},warmups:1,iterations:10}}'])
    assert not error_code



def test_option_write(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['write', file])
    assert not error_code


def test_option_read(file=generate_test_path('integers.ion')):
    # make sure it reads successfully
    (error_code, _, _) = run_cli(['read', file])
    assert not error_code


def test_option_write_no_c_extension(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['write', file, '--no-c-extension'])
    assert not error_code


def test_option_read_no_c_extension(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['read', file, '--no-c-extension'])
    assert not error_code


def test_option_read_iterations(file=generate_test_path('integers.ion')):
    # This is a potentially flaky test due to the overhead of running the CLI as a new process.
    start = time.perf_counter()
    (error_code, _, _) = run_cli(['read', file, '--iterations', '3'])
    stop = time.perf_counter()
    assert not error_code
    time_1 = stop - start

    start = time.perf_counter()
    (error_code, _, _) = run_cli(['read', file, '--iterations', '300'])
    stop = time.perf_counter()
    assert not error_code
    time_2 = stop - start

    assert time_2 > time_1


def test_option_write_iterations(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['write', file, '--iterations', '100'])
    assert not error_code


# Streaming not supported yet
# def test_read_multi_api(file=generate_test_path('integers.ion')):
#     execution_with_command(['read', file, '--api', 'load_dump', '--api', 'streaming'])

# Streaming not supported yet
# def test_write_multi_api(file=generate_test_path('integers.ion')):
#     execution_with_command(['write', file, '--api', 'load_dump', '--api', 'streaming'])


def test_read_duplicated_api(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['read', file, '--api', 'load_dump', '--api', 'load_dump'])
    assert not error_code


def test_write_duplicated_api(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['write', file, '--api', 'load_dump', '--api', 'load_dump'])
    assert not error_code


def test_read_multi_format(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['read', file, '--format', 'ion_text', '--format', 'ion_binary'])
    assert not error_code


def test_write_multi_format(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['write', file, '--format', 'ion_text', '--format', 'ion_binary'])
    assert not error_code


def test_read_multi_duplicated_format(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['read', file, '--format', 'ion_text', '--format', 'ion_binary', '--format', 'ion_text'])
    assert not error_code


def test_write_multi_duplicated_format(file=generate_test_path('integers.ion')):
    (error_code, _, _) = run_cli(['write', file, '--format', 'ion_text', '--format', 'ion_binary', '--format', 'ion_text', ])
    assert not error_code


@parametrize(
    *tuple((f.value for f in Format.Format if Format.format_is_json(f.value)))
)
def test_write_json_format(f):
    (error_code, _, _) = run_cli(['write', generate_test_path('json/object.json'), '--format', f'{f}'])
    assert not error_code


@parametrize(
    *tuple((f.value for f in Format.Format if Format.format_is_json(f.value)))
)
def test_read_json_format(f):
    (error_code, _, _) = run_cli(['read', generate_test_path('json/object.json'), '--format', f'{f}'])
    assert not error_code


@parametrize(
    *tuple((f.value for f in Format.Format if Format.format_is_cbor(f.value)))
)
def test_write_cbor_format(f):
    (error_code, _, _) = run_cli(['write', generate_test_path('cbor/sample'), '--format', f'{f}'])
    assert not error_code


@parametrize(
    *tuple((f.value for f in Format.Format if Format.format_is_cbor(f.value)))
)
def test_read_cbor_format(f):
    (error_code, _, _) = run_cli(['read', generate_test_path('cbor/sample'), '--format', f'{f}'])
    assert not error_code



@parametrize(*['buffer', 'file'])
def test_write_io_type(f):
    (error_code, _, _) = run_cli(['write', generate_test_path('integers.ion'), '--io-type', f'{f}', '--format', 'json'])
    assert not error_code



@parametrize(*['buffer', 'file'])
def test_read_io_type(f):
    (error_code, _, _) = run_cli(['read', '--io-type', f'{f}', '--format', 'ion_text', '--format', 'ion_binary', generate_test_path('integers.ion')])
    assert not error_code


@parametrize(
    *tuple((Format.Format.ION_TEXT, Format.Format.ION_BINARY))
)
def test_format_is_ion(f):
    assert format_is_ion(f.value) is True


@parametrize(
    *tuple((Format.Format.JSON,
            Format.Format.UJSON,
            Format.Format.RAPIDJSON,
            Format.Format.SIMPLEJSON
            ))
)
def test_format_is_json(f):
    assert format_is_json(f.value) is True


@parametrize(
    Format.Format.CBOR,
    Format.Format.CBOR2
)
def test_format_is_cbor(f):
    assert format_is_cbor(f.value) is True


def assert_ion_string_equals(act, exp):
    assert ion_equals(simpleion.loads(act), simpleion.loads(exp))


def test_compare_without_regression():
    (error_code, _, _) = run_cli(['compare', generate_test_path('compare/cats_baseline.ion'), generate_test_path('compare/cats_baseline.ion'), '--fail'])
    assert not error_code


def test_compare_with_small_regression():
    (error_code, _, _) = run_cli(['compare', generate_test_path('compare/cats_baseline.ion'), generate_test_path('compare/cats_small_regression.ion'), '--fail'])
    assert not error_code


def test_compare_with_large_regression():
    (error_code, _, _) = run_cli(['compare', './tests/benchmark_sample_data/compare/cats_baseline.ion', generate_test_path('compare/cats_large_regression.ion'), '--fail'])
    assert error_code


def test_format_conversion_ion_binary_to_ion_text():
    rewrite_file_to_format(generate_test_path('integers.ion'), Format.Format.ION_BINARY.value)
    assert os.path.exists('temp_integers.10n')
    os.remove('temp_integers.10n')


def test_format_conversion_ion_text_to_ion_binary():
    rewrite_file_to_format(generate_test_path('integers.10n'), Format.Format.ION_TEXT.value)
    assert os.path.exists('temp_integers.ion')
    os.remove('temp_integers.ion')



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from tests import parametrize

from amazon.ion.writer_binary_raw_fields import _write_int, _write_int_uncached, _write_varint, \
                                                _write_varint_uncached, _write_uint, \
                                                _write_uint_uncached, _write_varuint, \
                                                _write_varuint_uncached, _CACHE_SIZE


class WriteParameter:
    def __init__(self, value, expected, methods):
        self.value = value
        self.expected = expected
        self.methods = methods

    def __str__(self):
        return "[" + str(self.value) + ", " + str(self.expected) + ", " + str(self.methods) + "]"


int_methods = (_write_int, _write_int_uncached)
varint_methods = (_write_varint, _write_varint_uncached)
uint_methods = (_write_uint, _write_uint_uncached)
varuint_methods = (_write_varuint, _write_varuint_uncached)


def assert_value(p):
    expected = bytearray(p.expected)
    for method in p.methods:
        out = bytearray()
        method(out, p.value)
        assert expected == out


@parametrize(
    WriteParameter(0, [0x00], int_methods),
    WriteParameter(0, [0x00], uint_methods),
    WriteParameter(0, [0x80], varint_methods),
    WriteParameter(0, [0x80], varuint_methods),
    WriteParameter(1, [0x01], int_methods),
    WriteParameter(1, [0x01], uint_methods),
    WriteParameter(1, [0x81], varint_methods),
    WriteParameter(1, [0x81], varuint_methods),
    WriteParameter(-1, [0x81], int_methods),
    WriteParameter(-1, [0xC1], varint_methods),
)
def test_write_basic(p):
    assert_value(p)


@parametrize(
    WriteParameter(126, [0x7E], int_methods),
    WriteParameter(127, [0x7F], int_methods),
    WriteParameter(128, [0x00, 0x80], int_methods),
    WriteParameter(32766, [0x7F, 0xFE], int_methods),
    WriteParameter(32767, [0x7F, 0xFF], int_methods),
    WriteParameter(32768, [0x00, 0x80, 0x00], int_methods),
    WriteParameter(8388606, [0x7F, 0xFF, 0xFE], int_methods),
    WriteParameter(8388607, [0x7F, 0xFF, 0xFF], int_methods),
    WriteParameter(8388608, [0x00, 0x80, 0x00, 0x00], int_methods),
    WriteParameter(-126, [0xFE], int_methods),
    WriteParameter(-127, [0xFF], int_methods),
    WriteParameter(-128, [0x80, 0x80], int_methods),
    WriteParameter(-32766, [0xFF, 0xFE], int_methods),
    WriteParameter(-32767, [0xFF, 0xFF], int_methods),
    WriteParameter(-32768, [0x80, 0x80, 0x00], int_methods),
    WriteParameter(-8388606, [0xFF, 0xFF, 0xFE], int_methods),
    WriteParameter(-8388607, [0xFF, 0xFF, 0xFF], int_methods),
    WriteParameter(-8388608, [0x80, 0x80, 0x00, 0x00], int_methods),
    WriteParameter(62, [0xBE], varint_methods),
    WriteParameter(63, [0xBF], varint_methods),
    WriteParameter(64, [0x00, 0xC0], varint_methods),
    WriteParameter(8190, [0x3F, 0xFE], varint_methods),
    WriteParameter(8191, [0x3F, 0xFF], varint_methods),
    WriteParameter(8192, [0x00, 0x40, 0x80], varint_methods),
    WriteParameter(1048574, [0x3F, 0x7F, 0xFE], varint_methods),
    WriteParameter(1048575, [0x3F, 0x7F, 0xFF], varint_methods),
    WriteParameter(1048576, [0x00, 0x40, 0x00, 0x80], varint_methods),
    WriteParameter(-62, [0xFE], varint_methods),
    WriteParameter(-63, [0xFF], varint_methods),
    WriteParameter(-64, [0x40, 0xC0], varint_methods),
    WriteParameter(-8190, [0x7F, 0xFE], varint_methods),
    WriteParameter(-8191, [0x7F, 0xFF], varint_methods),
    WriteParameter(-8192, [0x40, 0x40, 0x80], varint_methods),
    WriteParameter(-1048574, [0x7F, 0x7F, 0xFE], varint_methods),
    WriteParameter(-1048575, [0x7F, 0x7F, 0xFF], varint_methods),
    WriteParameter(-1048576, [0x40, 0x40, 0x00, 0x80], varint_methods),
    WriteParameter(254, [0xFE], uint_methods),
    WriteParameter(255, [0xFF], uint_methods),
    WriteParameter(256, [0x01, 0x00], uint_methods),
    WriteParameter(65534, [0xFF, 0xFE], uint_methods),
    WriteParameter(65535, [0xFF, 0xFF], uint_methods),
    WriteParameter(65536, [0x01, 0x00, 0x00], uint_methods),
    WriteParameter(16777214, [0xFF, 0xFF, 0xFE], uint_methods),
    WriteParameter(16777215, [0xFF, 0xFF, 0xFF], uint_methods),
    WriteParameter(16777216, [0x01, 0x00, 0x00, 0x00], uint_methods),
    WriteParameter(126, [0xFE], varuint_methods),
    WriteParameter(127, [0xFF], varuint_methods),
    WriteParameter(128, [0x01, 0x80], varuint_methods),
    WriteParameter(16382, [0x7F, 0xFE], varuint_methods),
    WriteParameter(16383, [0x7F, 0xFF], varuint_methods),
    WriteParameter(16384, [0x01, 0x00, 0x80], varuint_methods),
    WriteParameter(2097150, [0x7F, 0x7F, 0xFE], varuint_methods),
    WriteParameter(2097151, [0x7F, 0x7F, 0xFF], varuint_methods),
    WriteParameter(2097152, [0x01, 0x00, 0x00, 0x80], varuint_methods),
)
def test_value_boundaries(p):
    assert_value(p)


@parametrize(
    int_methods,
    varint_methods
)
def test_boundaries_signed(p):
    half = _CACHE_SIZE // 2
    assert_cached_values(p, range(-half - 1, half))  # Go just past the cache boundaries.


@parametrize(
    uint_methods,
    varuint_methods
)
def test_cache_unsigned(p):
    assert_cached_values(p, range(_CACHE_SIZE + 1))


def assert_cached_values(p, value_range):
    for i in value_range:
        cached_value = []
        uncached_value = []
        p[0](cached_value, i)
        p[1](uncached_value, i)
        assert cached_value == uncached_value



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from io import BytesIO

from pytest import raises

from amazon.ion.writer_buffer import BufferTree


def assert_buffer(buf, expected=b'1234'):
    out = BytesIO()
    for partial in buf.drain():
        out.write(partial)
    assert expected == out.getvalue()


def test_scalars():
    buf = BufferTree()
    buf.add_scalar_value(b'1')
    buf.add_scalar_value(b'2')
    buf.add_scalar_value(b'34')
    assert_buffer(buf)


def test_container():
    buf = BufferTree()
    buf.start_container()
    buf.add_scalar_value(b'34')
    buf.end_container(b'12')
    assert_buffer(buf)


def test_nested_containers():
    buf = BufferTree()
    buf.start_container()
    buf.add_scalar_value(b'2')
    buf.start_container()
    buf.add_scalar_value(b'4')
    buf.end_container(b'3')
    buf.end_container(b'1')
    assert_buffer(buf)


def test_nested_containers_at_same_start():
    buf = BufferTree()
    buf.start_container()
    buf.start_container()
    buf.start_container()
    buf.add_scalar_value(b'4')
    buf.end_container(b'3')
    buf.end_container(b'2')
    buf.end_container(b'1')
    assert_buffer(buf)


def test_scalar_after_empty_container():
    buf = BufferTree()
    buf.start_container()
    buf.add_scalar_value(b'2')
    buf.start_container()
    buf.end_container(b'3')
    buf.add_scalar_value(b'4')
    buf.end_container(b'1')
    assert_buffer(buf)


def test_container_length():
    buf = BufferTree()
    buf.start_container()
    buf.add_scalar_value(b'234')
    assert 3 ==  buf.current_container_length
    buf.end_container(b'1')
    assert 4 == buf.current_container_length
    assert_buffer(buf)


def test_reuse_with_scalars():
    buf = BufferTree()
    buf.add_scalar_value(b'1')
    buf.add_scalar_value(b'2')
    buf.add_scalar_value(b'34')
    assert_buffer(buf)  # drain resets the writer buffer
    buf.add_scalar_value(b'5')
    buf.add_scalar_value(b'6')
    buf.add_scalar_value(b'78')
    assert_buffer(buf, b'5678')


def test_reuse_with_containers():
    buf = BufferTree()
    buf.start_container()
    buf.add_scalar_value(b'2')
    buf.start_container()
    buf.add_scalar_value(b'4')
    buf.end_container(b'3')
    buf.end_container(b'1')
    assert_buffer(buf)
    buf.start_container()
    buf.add_scalar_value(b'78')
    buf.end_container(b'56')
    assert_buffer(buf, b'5678')


def test_drain_with_active_container_fails():
    buf = BufferTree()
    buf.start_container()
    buf.add_scalar_value(b'1')
    with raises(ValueError):
        for partial in buf.drain():
            pass


def test_end_container_at_depth_zero_fails():
    buf = BufferTree()
    buf.start_container()
    buf.add_scalar_value(b'2')
    buf.end_container(b'1')
    with raises(ValueError):
        buf.end_container(b'0')



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from io import BytesIO
from functools import partial

from tests import parametrize
from tests.trampoline_util import always_self, moves_to, yields_iter
from tests.trampoline_util import trampoline_scaffold, TrampolineParameters

from amazon.ion.core import DataEvent, Transition
from amazon.ion.core import ION_VERSION_MARKER_EVENT
from amazon.ion.writer import writer_trampoline, blocking_writer, partial_transition
from amazon.ion.writer import WriteEventType


# Trivial data for write events.
_TRIVIAL_DATA = b'DATA'


# Generates trivial write events.
def _trivial_event(write_event_type):
    return DataEvent(
        write_event_type,
        _TRIVIAL_DATA
    )

_PARTIAL_EVENT = _trivial_event(WriteEventType.HAS_PENDING)
_COMPLETE_EVENT = _trivial_event(WriteEventType.COMPLETE)
_NEEDS_INPUT_EVENT = _trivial_event(WriteEventType.NEEDS_INPUT)


def _trivial_transition(event_type, data, delegate):
    return Transition(DataEvent(event_type, data), delegate)


_partial_result = partial(partial_transition, _TRIVIAL_DATA)
_complete_result = partial(_trivial_transition, WriteEventType.COMPLETE, _TRIVIAL_DATA)
_needs_input_result = partial(_trivial_transition, WriteEventType.NEEDS_INPUT, _TRIVIAL_DATA)


_P = TrampolineParameters


@parametrize(
    _P(
        desc='START WITH NONE',
        coroutine=always_self(_partial_result),
        input=[None],
        expected=[TypeError],
    ),
    _P(
        desc='ALWAYS PARTIAL SELF - NORMAL',
        coroutine=always_self(_partial_result),
        input=[ION_VERSION_MARKER_EVENT] + [None] * 3,
        expected=[_PARTIAL_EVENT] * 4
    ),
    _P(
        desc='ALWAYS PARTIAL SELF - NON-NONE',
        coroutine=always_self(_partial_result),
        input=[ION_VERSION_MARKER_EVENT, ION_VERSION_MARKER_EVENT],
        expected=[_PARTIAL_EVENT, TypeError]
    ),
    _P(
        desc='ALWAYS COMPLETE SELF - NORMAL',
        coroutine=always_self(_complete_result),
        input=[ION_VERSION_MARKER_EVENT] * 10,
        expected=[_COMPLETE_EVENT] * 10,
    ),
    _P(
        desc='ALWAYS COMPLETE SELF - NONE',
        coroutine=always_self(_complete_result),
        input=[ION_VERSION_MARKER_EVENT, None],
        expected=[_COMPLETE_EVENT, TypeError],
    ),
    _P(
        desc='ALWAYS NEEDS INPUT SELF - NORMAL',
        coroutine=always_self(_needs_input_result),
        input=[ION_VERSION_MARKER_EVENT] * 10,
        expected=[_NEEDS_INPUT_EVENT] * 10,
    ),
    _P(
        desc='ALWAYS NEEDS INPUT SELF - NONE',
        coroutine=always_self(_needs_input_result),
        input=[ION_VERSION_MARKER_EVENT, None],
        expected=[_NEEDS_INPUT_EVENT, TypeError],
    ),
    _P(
        desc='MOVES TO',
        coroutine=moves_to(always_self(_complete_result), _needs_input_result),
        input=[ION_VERSION_MARKER_EVENT] * 4,
        expected=[_NEEDS_INPUT_EVENT] + [_COMPLETE_EVENT] * 3,
    ),
)
def test_trampoline(p):
    trampoline_scaffold(writer_trampoline, p)


@parametrize(
    _P(
        desc='SINGLE COMPLETE EVENT',
        coroutine=yields_iter(_COMPLETE_EVENT),
        input=1,
        expected=_TRIVIAL_DATA
    ),
    _P(
        desc='MULTIPLE NEEDS INPUT EVENT',
        coroutine=yields_iter(*([_NEEDS_INPUT_EVENT] * 4)),
        input=4,
        expected=_TRIVIAL_DATA * 4
    ),
    _P(
        desc='PARTIAL THEN COMPLETE EVENT',
        coroutine=yields_iter(_PARTIAL_EVENT, _COMPLETE_EVENT),
        input=1,
        expected=_TRIVIAL_DATA * 2
    ),
)
def test_blocking_writer(p):
    buf = BytesIO()
    writer = blocking_writer(p.coroutine, buf)
    for i in range(p.input):
        result_type = writer.send(None)
        assert isinstance(result_type, WriteEventType) and result_type is not WriteEventType.HAS_PENDING
    assert p.expected == buf.getvalue()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from io import BytesIO

from pytest import raises

from amazon.ion.core import IonEvent, IonEventType, IonType
from amazon.ion.exceptions import IonException
from amazon.ion.symbols import SymbolTable, LOCAL_TABLE_TYPE
from amazon.ion.writer import blocking_writer
from amazon.ion.writer_binary import binary_writer


def new_writer(imports=None):
    out = BytesIO()
    return out, blocking_writer(binary_writer(imports), out)


def finish(writer):
    writer.send(IonEvent(IonEventType.STREAM_END))


def test_no_ivm_without_values():
    out, writer = new_writer()
    finish(writer)
    buf = out.getvalue()
    assert len(buf) == 0


def assert_ivm(buf, count=1):
    ivm = BytesIO()
    ivm.write(bytearray([0xE0, 0x01, 0x00, 0xEA]))
    ivm_bytes = ivm.getvalue()
    assert buf.count(ivm_bytes) == count


def test_implicit_ivm():
    out, writer = new_writer()
    writer.send(IonEvent(IonEventType.SCALAR, IonType.NULL))  # This implicitly writes the IVM.
    finish(writer)
    buf = out.getvalue()
    assert_ivm(buf)


def test_one_ivm_per_stream():
    out, writer = new_writer()
    # Explicitly write the IVM.
    writer.send(IonEvent(IonEventType.VERSION_MARKER))
    # If the IVM not explicitly written, this would write one.
    writer.send(IonEvent(IonEventType.SCALAR, IonType.NULL))
    finish(writer)
    buf = out.getvalue()
    assert_ivm(buf)


def test_finish_resets_buffers():
    out, writer = new_writer()
    writer.send(IonEvent(IonEventType.SCALAR, IonType.STRING, u'foo'))
    finish(writer)
    buf = out.getvalue()
    assert len(buf) != 0
    out.truncate(0)
    finish(writer)
    buf = out.getvalue()
    assert len(buf) == 0


def test_import_lst_fails():
    lst = SymbolTable(LOCAL_TABLE_TYPE, [u'foo'])
    out, writer = new_writer([lst])
    with raises(IonException):  # TODO fail earlier?
        writer.send(IonEvent(IonEventType.SCALAR, IonType.STRING, u'bar'))


def test_reuse_after_flush():
    out, writer = new_writer()
    writer.send(IonEvent(IonEventType.SCALAR, IonType.BOOL, True))
    finish(writer)
    length = len(out.getvalue())
    writer.send(IonEvent(IonEventType.SCALAR, IonType.BOOL, False))
    finish(writer)
    buf = out.getvalue()
    assert length != len(buf)
    assert_ivm(buf, 2)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Callable

from dataclasses import dataclass

from amazon.ionbenchmark.benchmark_runner import BenchmarkResult
from amazon.ionbenchmark.benchmark_spec import BenchmarkSpec


@dataclass(frozen=True)
class ReportField:
    """
    Represents a field that can be included in the benchmark test report
    """
    name: str
    compute_fn: Callable  # Callable[[BenchmarkSpec, BenchmarkResult], Any]
    units: str = None
    # Direction of improvement, used by the compare command
    doi: int = None


REPORT_FIELDS = [
    # TODO: Make sure we have the fields we need to perform a statistically meaningful comparison
    # I.e. if we end up needing to use ANOVA or Independent Samples T Test, do we have the fields we need?
    ReportField(name="format",
                compute_fn=lambda spec, _: spec.get_format()),
    ReportField(name="input_file",
                compute_fn=lambda spec, _: spec.get_input_file()),
    ReportField(name="operation",
                compute_fn=lambda spec, _: spec.derive_operation_name()),
    ReportField(name="file_size", units="B", doi=-1,
                compute_fn=lambda spec, _: spec.get_input_file_size()),
    ReportField(name="memory_usage_peak", units="B", doi=-1,
                compute_fn=lambda _, result: result.peak_memory_usage),
    ReportField(name="time_mean", units="ns", doi=-1,
                compute_fn=lambda _, result: result.nanos_per_op.mean),
    ReportField(name="time_min", units="ns", doi=-1,
                compute_fn=lambda _, result: result.nanos_per_op.min),
    ReportField(name="time_max", units="ns", doi=-1,
                compute_fn=lambda _, result: result.nanos_per_op.max),
    ReportField(name="time_sd", units="ns",
                compute_fn=lambda _, result: result.nanos_per_op.stdev),
    ReportField(name="time_rsd", units="%",
                compute_fn=lambda _, result: result.nanos_per_op.rstdev * 100),
    ReportField(name="time_error", units="ns",
                compute_fn=lambda _, result: result.nanos_per_op.margin_of_error(confidence=0.999)),
    ReportField(name="ops/s_mean", doi=+1,
                compute_fn=lambda _, result: result.ops_per_second.mean),
    ReportField(name="ops/s_min", doi=+1,
                compute_fn=lambda _, result: result.ops_per_second.min),
    ReportField(name="ops/s_max", doi=+1,
                compute_fn=lambda _, result: result.ops_per_second.max),
    ReportField(name="ops/s_sd",
                compute_fn=lambda _, result: result.ops_per_second.stdev),
    ReportField(name="ops/s_rsd", units="%",
                compute_fn=lambda _, result: result.ops_per_second.rstdev * 100),
    ReportField(name="ops/s_error",
                compute_fn=lambda _, result: result.ops_per_second.margin_of_error(confidence=0.999)),
]


def get_report_field_by_name(name: str):
    for field in REPORT_FIELDS:
        if name == field.name:
            return field
    raise ValueError(f"Not a valid report field: {name}")


def report_stats(benchmark_spec: BenchmarkSpec, benchmark_result: BenchmarkResult, report_fields: list):
    """
    Generate a report for the outcome of a running a benchmark.

    Available fields:
     * `name` – the name of the benchmark
     * `operation` – the operation performed by the benchmark ('load', 'loads', 'dump', 'dumps')
     * `file_size` – the size of the input data used in this benchmark
     * `input_file` – the file used for this benchmark
     * `format` – the format used for this benchmark
     * `memory_usage_peak` – the peak amount of memory allocated while running the benchmark function
     * `time_<stat>` – time statistic for the benchmark
     * `ops/s_<stat>` – number of operations (invocations of the benchmark function) per second

    `<stat>` can be `mean`, `min`, `max`, `median`, `error`, `stdev`, or `rstdev`

    :param benchmark_spec: The spec for the benchmark that was run
    :param benchmark_result: The output from the benchmark
    :param report_fields: list[str] of fields to include in the report.
    :return:
    """

    result = {'name': benchmark_spec.get_name()}

    for field_name in report_fields:
        field = get_report_field_by_name(field_name)
        if field.units is not None:
            key = f"{field.name}({field.units})"
        else:
            key = field.name
        result[key] = field.compute_fn(benchmark_spec, benchmark_result)

    return result



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from functools import partial
from io import BytesIO
from typing import NamedTuple, Coroutine, Sequence, Any

from pytest import raises

from tests import is_exception, parametrize
from tests.event_aliases import e_int
from tests.trampoline_util import always_self
from tests.trampoline_util import trampoline_scaffold

from amazon.ion.core import Transition
from amazon.ion.core import ION_VERSION_MARKER_EVENT
from amazon.ion.core import ION_STREAM_INCOMPLETE_EVENT, ION_STREAM_END_EVENT
from amazon.ion.reader import read_data_event, reader_trampoline, blocking_reader
from amazon.ion.reader import NEXT_EVENT, SKIP_EVENT
from amazon.ion.util import coroutine


_TRIVIAL_ION_EVENT = e_int(0)

_ivm_transition = partial(Transition, ION_VERSION_MARKER_EVENT)
_incomplete_transition = partial(Transition, ION_STREAM_INCOMPLETE_EVENT)
_end_transition = partial(Transition, ION_STREAM_END_EVENT)
_event_transition = partial(Transition, _TRIVIAL_ION_EVENT)


class ReaderTrampolineParameters(NamedTuple):
    desc: str
    coroutine: Coroutine
    input: Sequence[Any]
    expected: Sequence[Any]
    allow_flush: bool = False

    def __str__(self):
        return self.desc

_P = ReaderTrampolineParameters

_TRIVIAL_DATA_EVENT = read_data_event(b'DATA')
_EMPTY_DATA_EVENT = read_data_event(b'')


@parametrize(
    _P(
        desc='START WITH NONE',
        coroutine=always_self(_ivm_transition),
        input=[None],
        expected=[TypeError],
    ),
    _P(
        desc='START WITH SKIP',
        coroutine=always_self(_ivm_transition),
        input=[SKIP_EVENT],
        expected=[TypeError],
    ),
    _P(
        desc='ALWAYS IVM NEXT',
        coroutine=always_self(_ivm_transition),
        input=[NEXT_EVENT] * 4,
        expected=[ION_VERSION_MARKER_EVENT] * 4,
    ),
    _P(
        desc='ALWAYS IVM NEXT THEN SKIP',
        coroutine=always_self(_ivm_transition),
        input=[NEXT_EVENT, SKIP_EVENT],
        expected=[ION_VERSION_MARKER_EVENT, TypeError],
    ),
    _P(
        desc='ALWAYS INCOMPLETE NEXT EOF',
        coroutine=always_self(_incomplete_transition),
        input=[NEXT_EVENT, NEXT_EVENT],
        expected=[ION_STREAM_INCOMPLETE_EVENT, TypeError],
        allow_flush=False
    ),
    _P(
        desc='ALWAYS INCOMPLETE NEXT FLUSH',
        coroutine=always_self(_incomplete_transition),
        input=[NEXT_EVENT, NEXT_EVENT],
        expected=[ION_STREAM_INCOMPLETE_EVENT, ION_STREAM_INCOMPLETE_EVENT],
        allow_flush=True
    ),
    _P(
        desc='ALWAYS INCOMPLETE SKIP',
        coroutine=always_self(_incomplete_transition),
        input=[NEXT_EVENT, SKIP_EVENT],
        expected=[ION_STREAM_INCOMPLETE_EVENT, TypeError],
    ),
    _P(
        desc='ALWAYS INCOMPLETE DATA',
        coroutine=always_self(_incomplete_transition),
        input=[NEXT_EVENT] + [_TRIVIAL_DATA_EVENT] * 4,
        expected=[ION_STREAM_INCOMPLETE_EVENT] * 5,
    ),
    _P(
        desc='ALWAYS END NEXT',
        coroutine=always_self(_end_transition),
        input=[NEXT_EVENT, NEXT_EVENT],
        expected=[ION_STREAM_END_EVENT, TypeError],
    ),
    _P(
        desc='ALWAYS END SKIP',
        coroutine=always_self(_end_transition),
        input=[NEXT_EVENT, SKIP_EVENT],
        expected=[ION_STREAM_END_EVENT, TypeError],
    ),
    _P(
        desc='ALWAYS END DATA',
        coroutine=always_self(_end_transition),
        input=[NEXT_EVENT] + [_TRIVIAL_DATA_EVENT] * 4,
        expected=[ION_STREAM_END_EVENT] * 5,
    ),
    _P(
        desc='ALWAYS END EMPTY DATA',
        coroutine=always_self(_end_transition),
        input=[NEXT_EVENT] + [_EMPTY_DATA_EVENT],
        expected=[ION_STREAM_END_EVENT, ValueError],
    ),
    _P(
        desc='ALWAYS EVENT DATA',
        coroutine=always_self(_event_transition),
        input=[NEXT_EVENT] + [_TRIVIAL_DATA_EVENT],
        expected=[_TRIVIAL_ION_EVENT, TypeError]
    )
)
def test_trampoline(p):
    trampoline_scaffold(reader_trampoline, p, p.allow_flush)


class _P(NamedTuple):
    desc: str
    coroutine: Coroutine
    data: Any
    input: Any
    expected: Any

    def __str__(self):
        return self.desc


@coroutine
def _asserts_events(expecteds, outputs, allow_flush=False):
    output = None
    for expected, next_output in zip(expecteds, outputs):
        actual = yield output
        assert expected == actual
        output = next_output
    yield output
    if not allow_flush:
        raise EOFError()
    yield ION_STREAM_END_EVENT


@parametrize(
    _P(
        desc='ALWAYS COMPLETE',
        coroutine=_asserts_events(
            [NEXT_EVENT, SKIP_EVENT],
            [ION_VERSION_MARKER_EVENT] * 2),
        data=b'',
        input=[NEXT_EVENT, SKIP_EVENT],
        expected=[ION_VERSION_MARKER_EVENT] * 2
    ),
    _P(
        desc='FIRST INCOMPLETE, THEN COMPLETE',
        coroutine=_asserts_events(
            [NEXT_EVENT, read_data_event(b'a'), SKIP_EVENT],
            [ION_STREAM_INCOMPLETE_EVENT] + ([ION_VERSION_MARKER_EVENT] * 2)
        ),
        data=b'a',
        input=[NEXT_EVENT, SKIP_EVENT],
        expected=[ION_VERSION_MARKER_EVENT] * 2
    ),
    _P(
        desc='FIRST STREAM END, THEN COMPLETE',
        coroutine=_asserts_events(
            [NEXT_EVENT, read_data_event(b'a'), SKIP_EVENT],
            [ION_STREAM_END_EVENT] + ([ION_VERSION_MARKER_EVENT] * 2)
        ),
        data=b'a',
        input=[NEXT_EVENT, SKIP_EVENT],
        expected=[ION_VERSION_MARKER_EVENT] * 2
    ),
    _P(
        desc='PREMATURE EOF',
        coroutine=_asserts_events(
            [NEXT_EVENT, read_data_event(b'a'), read_data_event(b'b')],
            [ION_STREAM_END_EVENT, ION_STREAM_INCOMPLETE_EVENT, ION_STREAM_INCOMPLETE_EVENT],
            allow_flush=False
        ),
        data=b'ab',
        input=[NEXT_EVENT],
        expected=[EOFError]
    ),
    _P(
        desc='FLUSH',
        coroutine=_asserts_events(
            [NEXT_EVENT, read_data_event(b'a'), read_data_event(b'b')],
            [ION_STREAM_END_EVENT, ION_STREAM_INCOMPLETE_EVENT, ION_STREAM_INCOMPLETE_EVENT],
            allow_flush=True
        ),
        data=b'ab',
        input=[NEXT_EVENT],
        expected=[ION_STREAM_END_EVENT]
    ),
    _P(
        desc='SINGLE EVENT, THEN NATURAL EOF',
        coroutine=_asserts_events(
            [NEXT_EVENT, read_data_event(b'a'), NEXT_EVENT],
            [ION_STREAM_END_EVENT, ION_VERSION_MARKER_EVENT, ION_STREAM_END_EVENT]
        ),
        data=b'a',
        input=[NEXT_EVENT, NEXT_EVENT],
        expected=[ION_VERSION_MARKER_EVENT, ION_STREAM_END_EVENT]
    ),
)
def test_blocking_reader(p):
    buf = BytesIO(p.data)
    reader = blocking_reader(p.coroutine, buf, buffer_size=1)
    for input, expected in zip(p.input, p.expected):
        if is_exception(expected):
            with raises(expected):
                reader.send(input)
        else:
            actual = reader.send(input)
            assert expected == actual



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Provides support for Ion symbol tokens."""
import itertools
from itertools import chain
from itertools import islice
from itertools import repeat
from typing import NamedTuple, Optional

from .exceptions import CannotSubstituteTable

TEXT_ION = u'$ion'
TEXT_ION_1_0 = u'$ion_1_0'
TEXT_ION_SYMBOL_TABLE = u'$ion_symbol_table'
TEXT_NAME = u'name'
TEXT_VERSION = u'version'
TEXT_IMPORTS = u'imports'
TEXT_SYMBOLS = u'symbols'
TEXT_MAX_ID = u'max_id'
TEXT_ION_SHARED_SYMBOL_TABLE = u'$ion_shared_symbol_table'

SID_ION = 1
SID_ION_1_0 = 2
SID_ION_SYMBOL_TABLE = 3
SID_NAME = 4
SID_VERSION = 5
SID_IMPORTS = 6
SID_SYMBOLS = 7
SID_MAX_ID = 8
SID_ION_SHARED_SYMBOL_TABLE = 9


class ImportLocation(NamedTuple):
    """Represents the import location of a symbol token.

    An import location can be thought of as a position independent address of an imported symbol.
    This metadata, if defined, can indicate identity of a symbol that came from a shared symbol
    token that the application does not have access to.

    Note:
        Version is nor part of an import location because a position in a shared table by name
        uniquely identifies that slot irrespective of version.

    Args:
        name (unicode): The name of the shared symbol table.
        position (int): The position in the shared symbol table.
    """
    name: str
    position: int


class SymbolToken(NamedTuple):
    """Representation of a *symbolic token*.

    A symbolic token may be a *part* of an Ion value in several contexts:

    * The field name of a value in a ``struct``
    * An annotation of a value.
    * A scalar of type ``symbol`` in the Ion data model.

    Args:
        text (Optional[unicode]): The text image of the token.
        sid  (Optional[int]): The local symbol ID of the token.
        location (Optional[ImportLocation]): The import source of the token.

    Note:
        At least one of ``text`` or ``sid`` should be non-``None``
    """
    text: Optional[str]
    sid: Optional[int]
    location: Optional[ImportLocation] = None


def _system_symbol_token(text, sid):
    """Defines an Ion 1.0 system symbol token."""
    return SymbolToken(text, sid, ImportLocation(TEXT_ION, sid))


SYMBOL_ZERO_TOKEN = SymbolToken(None, 0)

_SYSTEM_SYMBOL_TOKENS = (
    _system_symbol_token(TEXT_ION, SID_ION),
    _system_symbol_token(TEXT_ION_1_0, SID_ION_1_0),
    _system_symbol_token(TEXT_ION_SYMBOL_TABLE, SID_ION_SYMBOL_TABLE),
    _system_symbol_token(TEXT_NAME, SID_NAME),
    _system_symbol_token(TEXT_VERSION, SID_VERSION),
    _system_symbol_token(TEXT_IMPORTS, SID_IMPORTS),
    _system_symbol_token(TEXT_SYMBOLS, SID_SYMBOLS),
    _system_symbol_token(TEXT_MAX_ID, SID_MAX_ID),
    _system_symbol_token(TEXT_ION_SHARED_SYMBOL_TABLE, SID_ION_SHARED_SYMBOL_TABLE),
)


class _SymbolTableType(NamedTuple):
    """A set of flags indicating attributes of a symbol table.

    Args:
        is_system (bool): Whether or not the symbol table is a system table.
        is_shared (bool): Whether or not the symbol table is a is_shared table.
        is_local (bool): Whether or not the symbol table is is_local.
    """
    is_system: bool
    is_shared: bool
    is_local: bool

SYSTEM_TABLE_TYPE = _SymbolTableType(is_system=True, is_shared=True, is_local=False)
SHARED_TABLE_TYPE = _SymbolTableType(is_system=False, is_shared=True, is_local=False)
LOCAL_TABLE_TYPE = _SymbolTableType(is_system=False, is_shared=False, is_local=True)


class SymbolTable(object):
    """A collection of symbols that is ordered.

    Symbol tables are basically an Unicode string to integer interning table.

    A few things to consider about symbol tables:

    * System symbol tables never have imports and are shared symbol tables themselves.
    * Shared symbol tables never import the system symbol table.
    * Local symbol tables implicitly import the system symbol table.

    Shared symbol tables have tokens that always have a ``location`` attribute referring to themselves.
    Local symbol tables have tokens whose ``location`` attribute refers to **either** the shared symbol
    that it was imported from or ``None`` if the symbol was defined locally.

    Note:
        Shared symbol tables (which include system symbol tables) are immutable.
        Local symbol tables support interning as a mutable operation.
        The implementation doesn't enforce making properties read-only to enforce this invariant.

    Args:
        table_type (_SymbolTableType): The type of symbol table.
        symbols (Iterable[unicode]): The symbols text associated *locally* to this table.
        imports (Optional[Iterable[SymbolTable]]): The imports of the table.
        name (Optional[unicode]): The name of this table.  Required for shared symbol tables.
        version (Optional[int]): The version of this table.  Required for shared symbol tables.
        is_substitute (Optional[bool]): Whether or not this table is substituted.  A substituted symbol
                table is one that is not resolvable and has placeholder entries.
    """
    def __init__(self, table_type, symbols, name=None, version=None, imports=None, is_substitute=False):
        if table_type.is_system and imports is not None:
            raise ValueError('System tables cannot have imports')
        if table_type.is_shared and (name is None or version is None or version <= 0):
            raise ValueError('Shared symbol tables must have a name and >= 1 version')
        if table_type.is_local and (name is not None or version is not None):
            raise ValueError('Local symbol tables cannot have a name or version')
        if table_type.is_system and (name != TEXT_ION):
            raise ValueError('System symbol tables must be named "%s"' % TEXT_ION)
        if name is not None and not isinstance(name, str):
            raise TypeError('Shared symbol tables must have a unicode name: %r' % name)

        self.table_type = table_type
        self.name = name
        self.version = version
        self.imports = imports
        self.is_substitute = is_substitute
        self.max_id = 0
        self.__symbols = []
        self.__mapping = {}

        if table_type.is_local or table_type.is_system:
            for token in _SYSTEM_SYMBOL_TOKENS:
                self.__add(token)
            self.max_id = len(_SYSTEM_SYMBOL_TOKENS)

        if imports is not None:
            for table in imports:
                for token in table:
                    if table_type.is_shared:
                        self.__add_shared(token)
                    else:
                        if not table.table_type.is_local \
                                or token.location is None \
                                or token.location.name != TEXT_ION:
                            # TODO Determine if this code should handle LST as import.
                            # If the import is a local symbol table, we need to ignore system
                            # imports.  This supports LST append.
                            self.__add_import(token)

        # System symbols are bootstrapped
        if not table_type.is_system:
            for symbol in symbols:
                self.__add_text(symbol)

    def __import_location(self, sid):
        """Returns a location for this table's SID.

        Only meaningful for shared tables.
        """
        return ImportLocation(self.name, sid)

    def __new_sid(self):
        """Allocates a new local SID."""
        self.max_id += 1
        sid = self.max_id
        return sid

    def __add(self, token):
        """Unconditionally adds a token to the table."""
        self.__symbols.append(token)
        text = token.text
        if text is not None and text not in self.__mapping:
            self.__mapping[text] = token

    def __add_shared(self, original_token):
        """Adds a token, normalizing the SID and import reference to this table."""
        sid = self.__new_sid()
        token = SymbolToken(original_token.text, sid, self.__import_location(sid))
        self.__add(token)
        return token

    def __add_import(self, original_token):
        """Adds a token, normalizing only the SID"""
        sid = self.__new_sid()
        token = SymbolToken(original_token.text, sid, original_token.location)
        self.__add(token)
        return token

    def __add_text(self, text):
        """Adds the given Unicode text as a locally defined symbol."""
        if text is not None and not isinstance(text, str):
            raise TypeError('Local symbol definition must be a Unicode sequence or None: %r' % text)
        sid = self.__new_sid()
        location = None
        if self.table_type.is_shared:
            location = self.__import_location(sid)
        token = SymbolToken(text, sid, location)
        self.__add(token)
        return token

    def intern(self, text):
        """Interns the given Unicode sequence into the symbol table.

        Note:
            This operation is only valid on local symbol tables.

        Args:
            text (unicode): The target to intern.

        Returns:
            SymbolToken: The mapped symbol token which may already exist in the table.
        """
        if self.table_type.is_shared:
            raise TypeError('Cannot intern on shared symbol table')
        if not isinstance(text, str):
            raise TypeError('Cannot intern non-Unicode sequence into symbol table: %r' % text)

        token = self.get(text)
        if token is None:
            token = self.__add_text(text)
        return token

    def get(self, key, default=None):
        """Returns a token by text or local ID, with a default.

        A given text image may be associated with more than one symbol ID.  This will return the first definition.

        Note:
            User defined symbol IDs are always one-based.  Symbol zero is a special symbol that
            always has no text.

        Args:
            key (unicode | int):  The key to lookup.
            default(Optional[SymbolToken]): The default to return if the key is not found

        Returns:
            SymbolToken: The token associated with the key or the default if it doesn't exist.
        """
        if isinstance(key, str):
            return self.__mapping.get(key, None)
        if not isinstance(key, int):
            raise TypeError('Key must be int or Unicode sequence.')

        # TODO determine if $0 should be returned for all symbol tables.
        if key == 0:
            return SYMBOL_ZERO_TOKEN

        # Translate one-based SID to zero-based intern table
        index = key - 1
        if index < 0 or key > len(self):
            return default
        return self.__symbols[index]

    def __getitem__(self, key):
        """Returns a token by text or local ID.

        Args:
            key (unicode | int): The text or ID to lookup.

        Returns:
            SymbolToken: The token associated with the key.

        Raises:
            KeyError: If the key is not in the table.

        See Also:
            :meth:`get()`
        """
        token = self.get(key)
        if token is None:
            raise KeyError('No symbol for: %s' % key)
        return token

    def __len__(self):
        """Returns the number of symbols within the table."""
        return len(self.__symbols)

    def __iter__(self):
        """Iterator over the table's tokens.

        Returns:
            Iterable[SymbolToken]: The tokens in this table in defined order.
        """
        return iter(self.__symbols)

    def __eq__(self, other):
        """Compares two symbol tables together.

        Two symbol tables are considered equal if the underlying tokens are the same and the
        ``table_type``, ``name``, ``version``, and ``is_substitute`` attributes are defined and are equal.

        Note:
            This is implemented using ``getattr`` to allow duck-typed table implementations to compare.
            Any custom symbol table-like implementation should implement this method accordingly.

            Things that are not compared:

             * ``imports`` are never compared as they are denormalized into the tokens of the table.
             * ``max_id`` is never compared as its an invariant that it matches the iteration.
        """
        if self is other:
            return True

        # Compare relevant attributes
        if self.table_type != getattr(other, 'table_type', None):
            return False
        if self.name != getattr(other, 'name', None):
            return False
        if self.version != getattr(other, 'version', None):
            return False
        if self.is_substitute != getattr(other, 'is_substitute', None):
            return False

        # Compare tokens.
        other_iter = getattr(other, '__iter__')
        if not callable(other_iter):
            return False
        for token, other_token in itertools.zip_longest(self, other):
            if token != other_token:
                return False

        return True

    def __ne__(self, other):
        return not self == other

SYSTEM_SYMBOL_TABLE = SymbolTable(
    table_type=SYSTEM_TABLE_TYPE,
    symbols=_SYSTEM_SYMBOL_TOKENS,
    name=TEXT_ION,
    version=1
)


def local_symbol_table(imports=None, symbols=()):
    """Constructs a local symbol table.

    Args:
        imports (Optional[SymbolTable]): Shared symbol tables to import.
        symbols (Optional[Iterable[Unicode]]): Initial local symbols to add.

    Returns:
        SymbolTable: A mutable local symbol table with the seeded local symbols.
    """
    return SymbolTable(
        table_type=LOCAL_TABLE_TYPE,
        symbols=symbols,
        imports=imports
    )


def shared_symbol_table(name, version, symbols, imports=None):
    """Constructs a shared symbol table.

    Args:
        name (unicode): The name of the shared symbol table.
        version (int): The version of the shared symbol table.
        symbols (Iterable[unicode]): The symbols to associate with the table.
        imports (Optional[Iterable[SymbolTable]): The shared symbol tables to inject into this one.

    Returns:
        SymbolTable: The constructed table.
    """
    return SymbolTable(
        table_type=SHARED_TABLE_TYPE,
        symbols=symbols,
        name=name,
        version=version,
        imports=imports
    )


def placeholder_symbol_table(name, version, max_id):
    """Constructs a shared symbol table that consists symbols that all have no known text.

    This is generally used for cases where a shared symbol table is not available by the
    application.

    Args:
        name (unicode): The name of the shared symbol table.
        version (int): The version of the shared symbol table.
        max_id (int): The maximum ID allocated by this symbol table, must be ``>= 0``

    Returns:
        SymbolTable: The synthesized table.
    """
    if version <= 0:
        raise ValueError('Version must be grater than or equal to 1: %s' % version)
    if max_id < 0:
        raise ValueError('Max ID must be zero or positive: %s' % max_id)

    return SymbolTable(
        table_type=SHARED_TABLE_TYPE,
        symbols=repeat(None, max_id),
        name=name,
        version=version,
        is_substitute=True
    )


def substitute_symbol_table(table, version, max_id):
    """Substitutes a given shared symbol table for another version.

    * If the given table has **more** symbols than the requested substitute, then the generated
      symbol table will be a subset of the given table.
    * If the given table has **less** symbols than the requested substitute, then the generated
      symbol table will have symbols with unknown text generated for the difference.

    Args:
        table (SymbolTable): The shared table to derive from.
        version (int): The version to target.
        max_id (int): The maximum ID allocated by the substitute, must be ``>= 0``.

    Returns:
        SymbolTable: The synthesized table.
    """
    if not table.table_type.is_shared:
        raise ValueError('Symbol table to substitute from must be a shared table')
    if version <= 0:
        raise ValueError('Version must be grater than or equal to 1: %s' % version)
    if max_id < 0:
        raise ValueError('Max ID must be zero or positive: %s' % max_id)

    # TODO Recycle the symbol tokens from the source table into the substitute.
    if max_id <= table.max_id:
        symbols = (token.text for token in islice(table, max_id))
    else:
        symbols = chain(
            (token.text for token in table),
            repeat(None, max_id - table.max_id)
        )

    return SymbolTable(
        table_type=SHARED_TABLE_TYPE,
        symbols=symbols,
        name=table.name,
        version=version,
        is_substitute=True
    )


class SymbolTableCatalog(object):
    """A collection of symbol tables that can be used to resolve imports.

    Note:
        The catalog will return a placeholder symbol table when resolving a table
        that doesn't exist. For tables that don't exist with any version, this placeholder
        will be completely devoid of text mappings.  For tables that exist with a non-exact version,
        a derived substitute will be generated.
    """
    def __init__(self):
        self.__tables = {}

    def register(self, table):
        """Adds a shared table to the catalog.

        Args:
            table (SymbolTable): A non-system, shared symbol table.
        """
        if table.table_type.is_system:
            raise ValueError('Cannot add system table to catalog')
        if not table.table_type.is_shared:
            raise ValueError('Cannot add local table to catalog')
        if table.is_substitute:
            raise ValueError('Cannot add substitute table to catalog')

        versions = self.__tables.get(table.name)
        if versions is None:
            versions = {}
            self.__tables[table.name] = versions
        versions[table.version] = table

    def resolve(self, name, version, max_id):
        """Resolves the table for a given name and version.

        Args:
            name (unicode): The name of the table to resolve.
            version (int): The version of the table to resolve.
            max_id (Optional[int]): The maximum ID of the table requested.
                May be ``None`` in which case an exact match on ``name`` and ``version``
                is required.

        Returns:
            SymbolTable: The *closest* matching symbol table.  This is either an exact match,
            a placeholder, or a derived substitute depending on what tables are registered.
        """
        if not isinstance(name, str):
            raise TypeError('Name must be a Unicode sequence: %r' % name)
        if not isinstance(version, int):
            raise TypeError('Version must be an int: %r' % version)
        if version <= 0:
            raise ValueError('Version must be positive: %s' % version)
        if max_id is not None and max_id < 0:
            raise ValueError('Max ID must be zero or positive: %s' % max_id)

        versions = self.__tables.get(name)
        if versions is None:
            if max_id is None:
                raise CannotSubstituteTable(
                    'Found no table for %s, but no max_id' % name
                )
            return placeholder_symbol_table(name, version, max_id)

        table = versions.get(version)
        if table is None:
            # TODO Replace the keys map with a search tree based dictionary.
            keys = list(versions)
            keys.sort()
            table = versions[keys[-1]]

        if table.version == version and (max_id is None or table.max_id == max_id):
            return table

        if max_id is None:
            raise CannotSubstituteTable(
                'Found match for %s, but not version %d, and no max_id' % (name, version)
            )

        return substitute_symbol_table(table, version, max_id)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
from datetime import datetime, timedelta
from functools import partial
from io import BytesIO

from decimal import Decimal
from itertools import chain
from math import isnan

import re
from typing import NamedTuple, Any, Sequence, Optional

from pytest import raises

from amazon.ion import simpleion
from amazon.ion.exceptions import IonException
from amazon.ion.symbols import SymbolToken, SYSTEM_SYMBOL_TABLE
from amazon.ion.writer_binary import _IVM
from amazon.ion.core import IonType, IonEvent, IonEventType, OffsetTZInfo, Multimap, TimestampPrecision
from amazon.ion.simple_types import IonPyDict, IonPyText, IonPyList, IonPyNull, IonPyBool, IonPyInt, IonPyFloat, \
    IonPyDecimal, IonPyTimestamp, IonPyBytes, IonPySymbol
from amazon.ion.equivalence import ion_equals, obj_has_ion_type_and_annotation
from amazon.ion.simpleion import dump, dumps, load, loads, _ion_type, _FROM_ION_TYPE, _FROM_TYPE_TUPLE_AS_SEXP, \
    _FROM_TYPE
from amazon.ion.writer_binary_raw import _serialize_symbol, _write_length
from tests.writer_util import VARUINT_END_BYTE, ION_ENCODED_INT_ZERO, SIMPLE_SCALARS_MAP_BINARY, SIMPLE_SCALARS_MAP_TEXT
from tests import parametrize
from amazon.ion.simpleion import c_ext

_st = partial(SymbolToken, sid=None, location=None)


class _Parameter(NamedTuple):
    desc: str
    obj: Any
    expected: Any
    has_symbols: bool
    stream: bool = False
    tuple_as_sexp: bool = False

    def __str__(self):
        return self.desc


class _Expected(NamedTuple):
    binary: Sequence[bytes]
    text: Sequence[bytes]


def expected(binary, text):
    return _Expected((binary,), (text,))


def bytes_of(*args, **kwargs):
    return bytes(bytearray(*args, **kwargs))


_SIMPLE_CONTAINER_MAP = {
    IonType.LIST: (
        (
            [[], ],
            (expected(b'\xB0', b'[]'),)
        ),
        (
            [(), ],
            (expected(b'\xB0', b'[]'),)
        ),
        (
            [IonPyList.from_value(IonType.LIST, []), ],
            (expected(b'\xB0', b'[]'),)
        ),
        (
            [[0], ],
            (expected(
                bytes_of([
                    0xB0 | 0x01,  # Int value 0 fits in 1 byte.
                    ION_ENCODED_INT_ZERO
                ]),
                b'[0]'
            ),)
        ),
        (
            [(0,), ],
            (expected(
                bytes_of([
                    0xB0 | 0x01,  # Int value 0 fits in 1 byte.
                    ION_ENCODED_INT_ZERO
                ]),
                b'[0]'
            ),)
        ),
        (
            [IonPyList.from_value(IonType.LIST, [0]), ],
            (expected(
                bytes_of([
                    0xB0 | 0x01,  # Int value 0 fits in 1 byte.
                    ION_ENCODED_INT_ZERO
                ]),
                b'[0]'
            ),)
        ),
    ),
    IonType.SEXP: (
        (
            [IonPyList.from_value(IonType.SEXP, []), ],
            (expected(b'\xC0', b'()'),)
        ),
        (
            [IonPyList.from_value(IonType.SEXP, [0]), ],
            (expected(
                bytes_of([
                    0xC0 | 0x01,  # Int value 0 fits in 1 byte.
                    ION_ENCODED_INT_ZERO
                ]),
                b'(0)'
            ),)
        ),
        (
            [(), ],  # NOTE: the generators will detect this and set 'tuple_as_sexp' to True for this case.
            (expected(b'\xC0', b'()'),)
        )
    ),
    IonType.STRUCT: (
        (
            [{}, ],
            (expected(b'\xD0', b'{}'),)
        ),
        (
            [IonPyDict.from_value(IonType.STRUCT, {}), ],
            (expected(b'\xD0', b'{}'),)
        ),
        (
            [{u'': u''}, ],
            (expected(
                bytes_of([
                    0xDE,  # The lower nibble may vary. It does not indicate actual length unless it's 0.
                    VARUINT_END_BYTE | 2,  # Field name 10 and value 0 each fit in 1 byte.
                    VARUINT_END_BYTE | 10,
                    0x80  # Empty string
                ]),
                b"{'':\"\"}"
            ),
             expected(
                 bytes_of([
                     0xD2,
                     VARUINT_END_BYTE | 10,
                     0x80  # Empty string
                 ]),
                 b"{'':\"\"}"
             ),
            )
        ),
        (
            [{u'foo': 0}, ],
            (expected(
                bytes_of([
                    0xDE,  # The lower nibble may vary. It does not indicate actual length unless it's 0.
                    VARUINT_END_BYTE | 2,  # Field name 10 and value 0 each fit in 1 byte.
                    VARUINT_END_BYTE | 10,
                    ION_ENCODED_INT_ZERO
                ]),
                b"{foo:0}"
            ),
             expected(
                 bytes_of([
                     0xD2,
                     VARUINT_END_BYTE | 10,
                     ION_ENCODED_INT_ZERO
                 ]),
                 b"{foo:0}"
             ),
            )
        ),
        (
            [IonPyDict.from_value(IonType.STRUCT, {u'foo': 0}), ],
            (expected(
                bytes_of([
                    0xDE,  # The lower nibble may vary. It does not indicate actual length unless it's 0.
                    VARUINT_END_BYTE | 2,  # Field name 10 and value 0 each fit in 1 byte.
                    VARUINT_END_BYTE | 10,
                    ION_ENCODED_INT_ZERO
                ]),
                b"{foo:0}"
            ),
             expected(
                 bytes_of([
                     0xD2,
                     VARUINT_END_BYTE | 10,
                     ION_ENCODED_INT_ZERO
                 ]),
                 b"{foo:0}"
             ),
            )
        ),
    ),
}


def generate_scalars_binary(scalars_map, preceding_symbols=0):
    for ion_type, values in iter(scalars_map.items()):
        for native, expected in values:
            native_expected = expected
            has_symbols = False
            if native is None:
                # An un-adorned 'None' doesn't contain enough information to determine its Ion type
                native_expected = b'\x0f'
            elif ion_type is IonType.CLOB:
                # All bytes are treated as BLOBs unless wrapped by an _IonNature
                tid = expected[0] + 0x10  # increment upper nibble for clob -> blob; keep lower nibble
                native_expected = bytearray([tid]) + expected[1:]
            elif ion_type is IonType.SYMBOL and native is not None:
                has_symbols = True
            elif ion_type is IonType.STRING:
                # Encode all strings as symbols too.
                if c_ext:
                    symbol_expected = _serialize_symbol(
                        IonEvent(IonEventType.SCALAR, IonType.SYMBOL, SymbolToken(None, 10)))
                else:
                    symbol_expected = _serialize_symbol(
                        IonEvent(IonEventType.SCALAR, IonType.SYMBOL, SymbolToken(None, 10 + preceding_symbols)))
                yield _Parameter(IonType.SYMBOL.name + ' ' + native,
                                 IonPyText.from_value(IonType.SYMBOL, native, ()), symbol_expected, True)
            yield _Parameter('%s %s' % (ion_type.name, native), native, native_expected, has_symbols)
            wrapper = _FROM_ION_TYPE[ion_type].from_value(ion_type, native)
            yield _Parameter(repr(wrapper), wrapper, expected, has_symbols)


def generate_containers_binary(container_map, preceding_symbols=0):
    for ion_type, container in iter(container_map.items()):
        for test_tuple in container:
            obj = test_tuple[0]
            expecteds = test_tuple[1]
            final_expected = ()
            has_symbols = False
            tuple_as_sexp = False
            for one_expected in expecteds:
                one_expected = one_expected.binary
                for elem in obj:
                    if isinstance(elem, (dict, Multimap, IonPyDict)) and len(elem) > 0:
                        has_symbols = True
                    elif ion_type is IonType.SEXP and isinstance(elem, tuple):
                        tuple_as_sexp = True
                if has_symbols and preceding_symbols:
                    # we need to make a distinct copy that will contain an altered encoding
                    one_expected = []
                    for expected in one_expected:
                        expected = bytearray(expected)
                        field_sid = expected[-2] & (~VARUINT_END_BYTE)
                        expected[-2] = VARUINT_END_BYTE | (preceding_symbols + field_sid)
                        one_expected.append(expected)
                expected = bytearray()
                for e in one_expected:
                    expected.extend(e)
                final_expected += (expected,)
            yield _Parameter(repr(obj), obj, final_expected, has_symbols, True, tuple_as_sexp=tuple_as_sexp)


def generate_annotated_values_binary(scalars_map, container_map):
    for value_p in chain(generate_scalars_binary(scalars_map, preceding_symbols=2),
                         generate_containers_binary(container_map, preceding_symbols=2)):
        obj = value_p.obj
        if not obj_has_ion_type_and_annotation(obj):
            continue
        obj.ion_annotations = (_st(u'annot1'), _st(u'annot2'),)
        annot_length = 2  # 10 and 11 each fit in one VarUInt byte
        annot_length_length = 1  # 2 fits in one VarUInt byte

        final_expected = ()
        if isinstance(value_p.expected, (list, tuple)):
            expecteds = value_p.expected
        else:
            expecteds = (value_p.expected, )
        for one_expected in expecteds:
            value_length = len(one_expected)
            length_field = annot_length + annot_length_length + value_length
            wrapper = []
            _write_length(wrapper, length_field, 0xE0)

            if c_ext and obj.ion_type is IonType.SYMBOL and not isinstance(obj, IonPyNull) \
                    and not (hasattr(obj, 'sid') and (obj.sid is None or obj.sid < 10)):
                wrapper.extend([
                    VARUINT_END_BYTE | annot_length,
                    VARUINT_END_BYTE | 11,
                    VARUINT_END_BYTE | 12
                ])
            else:
                wrapper.extend([
                    VARUINT_END_BYTE | annot_length,
                    VARUINT_END_BYTE | 10,
                    VARUINT_END_BYTE | 11
                ])

            exp = bytearray(wrapper) + one_expected
            final_expected += (exp, )

        yield _Parameter(
            desc='ANNOTATED %s' % value_p.desc,
            obj=obj,
            expected=final_expected,
            has_symbols=True,
            stream=value_p.stream
        )


def _assert_symbol_aware_ion_equals(assertion, output):
    if ion_equals(assertion, output):
        return True
    if isinstance(assertion, SymbolToken):
        expected_token = assertion
        if assertion.text is None:
            assert assertion.sid is not None
            # System symbol IDs are mapped correctly in the text format.
            token = SYSTEM_SYMBOL_TABLE.get(assertion.sid)
            assert token is not None  # User symbols with unknown text won't be successfully read.
            expected_token = token
        return expected_token == output
    else:
        try:
            return isnan(assertion) and isnan(output)
        except TypeError:
            return False


def _dump_load_run(p, dumps_func, loads_func, binary):
    # test dump
    res = dumps_func(p.obj, binary=binary, sequence_as_stream=p.stream, tuple_as_sexp=p.tuple_as_sexp,
                     omit_version_marker=True)
    if isinstance(p.expected, (tuple, list)):
        expecteds = p.expected
    else:
        expecteds = (p.expected,)
    write_success = False
    for expected in expecteds:
        if not p.has_symbols:
            if binary:
                write_success = (_IVM + expected) == res or expected == res
            else:
                write_success = (b'$ion_1_0 ' + expected) == res or expected == res
        else:
            # The payload contains a LST. The value comes last, so compare the end bytes.
            write_success = expected == res[len(res) - len(expected):]
        if write_success:
            break
    if not write_success:
        raise AssertionError('Expected: %s , found %s' % (expecteds, res))
    # test load
    res = loads_func(res, single_value=(not p.stream))
    _assert_symbol_aware_ion_equals(p.obj, res)


def _simple_dumps(obj, *args, **kw):
    buf = BytesIO()
    dump(obj, buf, *args, **kw)
    return buf.getvalue()


def _simple_loads(data, *args, **kw):
    buf = BytesIO()
    buf.write(data)
    buf.seek(0)
    return load(buf, *args, **kw)


@parametrize(
    *tuple(chain(
        generate_scalars_binary(SIMPLE_SCALARS_MAP_BINARY),
        generate_containers_binary(_SIMPLE_CONTAINER_MAP),
        generate_annotated_values_binary(SIMPLE_SCALARS_MAP_BINARY, _SIMPLE_CONTAINER_MAP),
    ))
)
def test_dump_load_binary(p):
    _dump_load_run(p, _simple_dumps, _simple_loads, binary=True)


@parametrize(
    *tuple(chain(
        generate_scalars_binary(SIMPLE_SCALARS_MAP_BINARY),
        generate_containers_binary(_SIMPLE_CONTAINER_MAP),
        generate_annotated_values_binary(SIMPLE_SCALARS_MAP_BINARY, _SIMPLE_CONTAINER_MAP),
    ))
)
def test_dumps_loads_binary(p):
    _dump_load_run(p, dumps, loads, binary=True)


def generate_scalars_text(scalars_map):
    for ion_type, values in iter(scalars_map.items()):
        for native, expected in values:
            native_expected = expected
            has_symbols = False
            if native is None:
                native_expected = b'null'
            elif ion_type is IonType.CLOB:
                # All bytes are treated as BLOBs unless wrapped by an _IonNature
                native = _FROM_ION_TYPE[ion_type].from_value(ion_type, native)
            elif ion_type is IonType.SYMBOL and native is not None:
                has_symbols = True
                if not isinstance(native, SymbolToken):
                    native = _st(native)
            yield _Parameter('%s %s' % (ion_type.name, native), native, native_expected, has_symbols)
            if not (ion_type is IonType.CLOB):
                # Clobs were already wrapped.
                wrapper = _FROM_ION_TYPE[ion_type].from_value(ion_type, native)
                yield _Parameter(repr(wrapper), wrapper, expected, has_symbols)


def generate_containers_text(container_map):
    for ion_type, container in iter(container_map.items()):
        for test_tuple in container:
            obj = test_tuple[0]
            expected = test_tuple[1]
            final_expected = ()
            has_symbols = False
            tuple_as_sexp = False

            for one_expected in expected:
                one_expected = one_expected.text[0]
                for elem in obj:
                    if isinstance(elem, dict) and len(elem) > 0:
                        has_symbols = True
                    elif ion_type is IonType.SEXP and isinstance(elem, tuple):
                        tuple_as_sexp = True
                final_expected += (one_expected,)
            yield _Parameter(repr(obj), obj, final_expected, has_symbols, True, tuple_as_sexp=tuple_as_sexp)


def generate_annotated_values_text(scalars_map, container_map):
    for value_p in chain(generate_scalars_text(scalars_map),
                         generate_containers_text(container_map)):
        obj = value_p.obj
        if not obj_has_ion_type_and_annotation(obj):
            continue
        obj.ion_annotations = (_st(u'annot1'), _st(u'annot2'),)

        annotated_expected = ()
        if isinstance(value_p.expected, (tuple, list)):
            for expected in value_p.expected:
                annotated_expected += (b"annot1::annot2::" + expected,)
        else:
            annotated_expected += (b"annot1::annot2::" + value_p.expected,)

        yield _Parameter(
            desc='ANNOTATED %s' % value_p.desc,
            obj=obj,
            expected=annotated_expected,
            has_symbols=True,
            stream=value_p.stream
        )


@parametrize(
    *tuple(chain(
        generate_scalars_text(SIMPLE_SCALARS_MAP_TEXT),
        generate_containers_text(_SIMPLE_CONTAINER_MAP),
        generate_annotated_values_text(SIMPLE_SCALARS_MAP_TEXT, _SIMPLE_CONTAINER_MAP),
    ))
)
def test_dump_load_text(p):
    _dump_load_run(p, _simple_dumps, _simple_loads, binary=False)


@parametrize(
    *tuple(chain(
        generate_scalars_text(SIMPLE_SCALARS_MAP_TEXT),
        generate_containers_text(_SIMPLE_CONTAINER_MAP),
        generate_annotated_values_text(SIMPLE_SCALARS_MAP_TEXT, _SIMPLE_CONTAINER_MAP),
    ))
)
def test_dumps_loads_text(p):
    def dump_func(*args, **kw):
        sval = dumps(*args, **kw)
        # encode to UTF-8 bytes for comparisons
        return sval.encode('UTF-8')

    _dump_load_run(p, dump_func, loads, binary=False)


_ROUNDTRIPS = [
    None,
    IonPyNull.from_value(IonType.NULL, None),
    IonPyNull.from_value(IonType.BOOL, None),
    IonPyNull.from_value(IonType.INT, None),
    IonPyNull.from_value(IonType.FLOAT, None),
    IonPyNull.from_value(IonType.DECIMAL, None),
    IonPyNull.from_value(IonType.TIMESTAMP, None),
    IonPyNull.from_value(IonType.SYMBOL, None),
    IonPyNull.from_value(IonType.STRING, None),
    IonPyNull.from_value(IonType.CLOB, None),
    IonPyNull.from_value(IonType.BLOB, None),
    IonPyNull.from_value(IonType.LIST, None),
    IonPyNull.from_value(IonType.SEXP, None),
    IonPyNull.from_value(IonType.STRUCT, None),
    True,
    False,
    IonPyInt.from_value(IonType.BOOL, 0),
    IonPyInt.from_value(IonType.BOOL, 1),
    0,
    -1,
    1.23,
    1.23e4,
    -1.23,
    -1.23e-4,
    Decimal(0),
    Decimal('-1.23'),
    datetime(year=1, month=1, day=1, tzinfo=OffsetTZInfo(timedelta(minutes=-1))),
    u'',
    u'abc',
    u'abcdefghijklmno',
    u'a\U0001f4a9c',
    u'a\u3000\x20c',
    _st(u''),
    _st(u'abc'),
    _st(u'abcdefghijklmno'),
    _st(u'a\U0001f4a9c'),
    _st(u'a\u3000\x20c'),
    b'abcd',
    IonPyBytes.from_value(IonType.CLOB, b'abcd'),
    [[[]]],
    [[], [], []],
    [{}, {}, {}],
    [(), (), ()],
    (((),),),
    ([], [], [],),
    {u'foo': [], u'bar': (), u'baz': []},
    {u'foo': {u'foo': {}}},
    [{u'foo': [{u'foo': []}]}],
    {u'foo': ({u'foo': []},)},
    {
         u'foo': IonPyText.from_value(IonType.STRING, u'bar', annotations=(u'str',)),
         u'baz': 123,
         u'lst': IonPyList.from_value(IonType.LIST, [
             True, None, 1.23e4, IonPyText.from_value(IonType.SYMBOL, u'sym')
         ]),
         u'sxp': IonPyList.from_value(IonType.SEXP, [
             False, IonPyNull.from_value(IonType.STRUCT, None, (u'class',)), Decimal('5.678'),
             IonPyText.from_value(IonType.SYMBOL, u'sym2'), IonPyText.from_value(IonType.SYMBOL, u'_a_s_d_f_')
         ]),
         u'lst_or_sxp': (123, u'abc')
    },

]


def _generate_roundtrips(roundtrips):
    for is_binary in (True, False):
        for indent in ('not used',) if is_binary else (None, '', ' ', '   ', '\t', '\n\t\n  '):
            def _adjust_sids(annotations=()):
                if is_binary and isinstance(obj, SymbolToken):
                    return SymbolToken(obj.text, 10 + len(annotations))
                return obj

            def _to_obj(to_type=None, annotations=(), tuple_as_sexp=False):
                if to_type is None:
                    to_type = ion_type
                obj_out = _adjust_sids(annotations)
                return _FROM_ION_TYPE[ion_type].from_value(to_type, obj_out, annotations=annotations), is_binary, indent, tuple_as_sexp

            for obj in roundtrips:
                obj = _adjust_sids()
                yield obj, is_binary, indent, False
                if not obj_has_ion_type_and_annotation(obj):
                    ion_type = _ion_type(obj, _FROM_TYPE)
                    yield _to_obj()
                else:
                    ion_type = obj.ion_type
                if isinstance(obj, IonPyNull):
                    obj = None
                yield _to_obj(annotations=(u'annot1', u'annot2'))
                if isinstance(obj, list):
                    yield _to_obj(IonType.SEXP)
                    yield _to_obj(IonType.SEXP, annotations=(u'annot1', u'annot2'))
                if isinstance(obj, tuple) and not isinstance(obj, SymbolToken):
                    yield _to_obj(IonType.SEXP, tuple_as_sexp=True)
                    yield _to_obj(IonType.SEXP, annotations=(u'annot1', u'annot2'), tuple_as_sexp=True)


def _assert_roundtrip(before, after, tuple_as_sexp):
    assert ion_equals(before, after)


@parametrize(
    *tuple(_generate_roundtrips(_ROUNDTRIPS))
)
def test_roundtrip(p):
    obj, is_binary, indent, tuple_as_sexp = p
    out = BytesIO()
    dump(obj, out, binary=is_binary, indent=indent, tuple_as_sexp=tuple_as_sexp)
    out.seek(0)
    res = load(out)
    _assert_roundtrip(obj, res, tuple_as_sexp)


@parametrize(
    *tuple(_generate_roundtrips(_ROUNDTRIPS))
)
def test_roundtrip_ion_stream(p):
    obj, is_binary, indent, tuple_as_sexp = p
    expected = [obj]
    out = BytesIO()
    dump(obj, out, binary=is_binary, indent=indent, tuple_as_sexp=tuple_as_sexp)
    out.seek(0)
    res = load(out, single_value=False, parse_eagerly=True)
    _assert_roundtrip(expected, res, tuple_as_sexp)


def test_ion_stream():
    expected = "$ion_1_0 [1] (2) {a:3}"
    result = dumps(loads(expected, single_value=False, parse_eagerly=False), binary=False)
    assert result == expected

@parametrize(True, False)
def test_single_value_with_stream_fails(is_binary):
    out = BytesIO()
    dump(['foo', 123], out, binary=is_binary, sequence_as_stream=True)
    out.seek(0)
    with raises(IonException):
        load(out, single_value=True)


@parametrize(True, False)
def test_unknown_object_type_fails(is_binary):
    class Dummy:
        pass
    out = BytesIO()
    with raises(TypeError):
        dump(Dummy(), out, binary=is_binary)


class PrettyPrintParams(NamedTuple):
    ion_text: str
    indent: str
    exact_text: Optional[str] = None
    regexes: Sequence = []

@parametrize(
        PrettyPrintParams(ion_text='a', indent='  ', exact_text="$ion_1_0\na"),
        PrettyPrintParams(ion_text='"a"', indent='  ', exact_text="$ion_1_0\n\"a\""),
        PrettyPrintParams(ion_text='\'$a__9\'', indent='  ', exact_text="$ion_1_0\n$a__9"),
        PrettyPrintParams(ion_text='\'$a_\\\'_9\'', indent='  ', exact_text="$ion_1_0\n\'$a_\\\'_9\'"),
        PrettyPrintParams(ion_text='[a, b, chair::2008-08-08T]', indent='  ',
            exact_text="$ion_1_0\n[\n  a,\n  b,\n  chair::2008-08-08T\n]"),
        PrettyPrintParams(ion_text='[a, b, chair::2008-08-08T]', indent=None, # not pretty print
            exact_text="$ion_1_0 [a,b,chair::2008-08-08T]"),
        PrettyPrintParams(ion_text='[apple, {roof: false}]', indent='\t',
            exact_text="$ion_1_0\n[\n\tapple,\n\t{\n\t\troof: false\n\t}\n]"),
        PrettyPrintParams(ion_text='[apple, "banana", {roof: false}]', indent='\t',
            exact_text="$ion_1_0\n[\n\tapple,\n\t\"banana\",\n\t{\n\t\troof: false\n\t}\n]"),
        PrettyPrintParams(ion_text='[apple, {roof: false, walls:4, door: wood::large::true}]', indent='\t',
            regexes=["\\A\\$ion_1_0\n\\[\n\tapple,\n\t\\{", "\n\t\tdoor: wood::large::true,?\n",
                "\n\t\troof: false,?\n", "\n\t\twalls: 4,?\n", "\n\t\\}\n\\]\\Z"])
        )
def test_pretty_print(p):
    if c_ext:
        # TODO support pretty print for C extension.
        return
    ion_text, indent, exact_text, regexes = p
    ion_value = loads(ion_text)
    actual_pretty_ion_text = dumps(ion_value, binary=False, indent=indent)
    if exact_text is not None:
        assert actual_pretty_ion_text == exact_text
    for regex_str in regexes:
        assert re.search(regex_str, actual_pretty_ion_text, re.M) is not None


# Regression test for issue #95
def test_struct_field():
    # pass a dict through simpleion to get a reconstituted dict of Ion values.
    struct_a = loads(dumps({u'dont_remember_my_name': 1}))

    # copy the value of the "dont_remember_my_name" field to a new struct, which is also passed through simpleion
    struct_b = {u'new_name': struct_a[u"dont_remember_my_name"]}
    struct_c = loads(dumps(struct_b))

    # The bug identified in ion-python#95 is that the name of the original field is somehow preserved.
    # verify this no longer happens
    assert u'dont_remember_my_name' not in struct_c
    assert u'new_name' in struct_c


def test_dumps_omit_version_marker():
    v = loads('5')
    assert dumps(v, binary=False) == '$ion_1_0 5'
    assert dumps(v, binary=False, omit_version_marker=True) == '5'

    # verify no impact on binary output
    assert dumps(v) == b'\xe0\x01\x00\xea\x21\x05'
    assert dumps(v, omit_version_marker=True) == b'\xe0\x01\x00\xea\x21\x05'


# See issue https://github.com/amazon-ion/ion-python/issues/213
def test_loads_unicode_utf8_conversion():
    # Generates test data that more than 1024*32 bytes
    data = "[ '''\u2013''',"
    data += 'test,' * 100000
    data += "]"
    # Loads API should convert it to UTF-8 without illegal bytes number read exception.
    loads(data, parse_eagerly=True)


# See issue https://github.com/amazon-ion/ion-python/issues/232
def test_loads_large_string():
    # This function only tests c extension
    if not c_ext:
        return

    data = "a"*100000

    # Without symbol_buffer_threshold setup, it should fail due to BUFFER_TOO_SMALL
    try:
        loads(data)
    except Exception:
        pass
    else:
        assert False

    # With symbol_buffer_threshold setup, it should have enough buffer size to handle "a"*100000
    try:
        loads(data, text_buffer_size_limit=200000)
    except Exception:
        assert False


# This test ensures that the c_ext flag does not override whether the extension is actually supported.
def test_setting_c_ext_flag():
    if not simpleion.c_ext:
        simpleion.c_ext = True

    value = loads("{a:1}")
    dumps(value)


@parametrize(
    (IonPyNull(), IonPyNull.from_value(IonType.NULL, None, ())),
    (IonPyDecimal(5.6), IonPyDecimal.from_value(IonType.DECIMAL, 5.6, ())),
    (IonPyBytes(b'asdf'), IonPyBytes.from_value(IonType.CLOB, b'asdf', ())),
    (IonPyInt(123), IonPyInt.from_value(IonType.INT, 123, ())),
    (IonPyBool(True), IonPyBool.from_value(IonType.BOOL, True, ())),
    (IonPyFloat(5.6), IonPyFloat.from_value(IonType.FLOAT, 5.6, ())),
    (IonPyText(u'string'), IonPyText.from_value(IonType.STRING, u'string', ())),
    (IonPyTimestamp(year=1, month=1, day=1, tzinfo=OffsetTZInfo(timedelta(minutes=-1))),
     IonPyTimestamp.from_value(IonType.TIMESTAMP,
                               datetime(year=1, month=1, day=1, tzinfo=OffsetTZInfo(timedelta(minutes=-1))),
                               ())),
    (IonPySymbol(text='symbol', sid=None, location=None), IonPySymbol.from_value(IonType.SYMBOL, 'symbol', ())),
    (IonPyList([]), IonPyList.from_value(IonType.LIST, [], ())),
    (IonPyDict({}), IonPyDict.from_value(IonType.STRUCT, {}, ())),
)
def test_ion_py_objects_construction(v):
    """These tests are related to the issue https://github.com/amazon-ion/ion-python/issues/297

    Note that the IonPyTimestamp constructor only sets the attributes for the base datetime type which does not have precision or fractional seconds. IonPyTimestamp.from_value sets these attributes after construction.
    So, we compare if they represent the same instant; in other words, we set timestamps_instants_only to True.
    """
    assert True is ion_equals(v[0], v[1], timestamps_instants_only=True)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Binary Ion writer with symbol table management."""
from enum import IntEnum
from typing import Optional, NamedTuple

from .core import ION_STREAM_END_EVENT, IonEventType, IonType, IonEvent, DataEvent, Transition, SymbolTokenOrStr
from .exceptions import IonException
from .symbols import SID_ION_SYMBOL_TABLE, SID_IMPORTS, SHARED_TABLE_TYPE, \
                     SID_NAME, SID_VERSION, SID_MAX_ID, SID_SYMBOLS, LOCAL_TABLE_TYPE, \
                     SymbolTable, _SYSTEM_SYMBOL_TOKENS
from .util import coroutine
from .writer import NOOP_WRITER_EVENT, writer_trampoline, partial_transition, WriteEventType, \
                    _drain
from .writer_binary_raw import _WRITER_EVENT_NEEDS_INPUT_EMPTY, _raw_binary_writer
from .writer_buffer import BufferTree

_IVM = b'\xE0\x01\x00\xEA'


class _SymbolEventType(IntEnum):
    """Enumeration of events that impact a symbol table.

    Attributes:
        START_LST: Indicates that a local symbol table should be started.
        SYMBOL: Indicates that a symbol should be interned into the local symbol,
            and the resulting symbol ID returned.
        FINISH: Indicates that the local symbol table (if any) should be ended.
    """
    START_LST = 0
    SYMBOL = 1
    FINISH = 2


class _SymbolEvent(NamedTuple):
    """Symbol event used by the managed writer coroutine to trigger an action in the symbol
    table coroutine.

    Only events of type SYMBOL should have symbol text associated with them.

    Args:
        event_type (_SymbolEventType): The type of symbol event.
        symbol (Optional[SymbolTokenOrStr]): The symbol token or text associated with the event.
    """
    event_type: _SymbolEventType
    symbol: Optional[SymbolTokenOrStr] = None


def _system_token(sid):
    return _SYSTEM_SYMBOL_TOKENS[sid - 1]

_SYMBOL_EVENT_START_LST = _SymbolEvent(_SymbolEventType.START_LST)
_SYMBOL_EVENT_FINISH = _SymbolEvent(_SymbolEventType.FINISH)

_ION_EVENT_STRUCT_START = IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT)
_ION_EVENT_STREAM_END = ION_STREAM_END_EVENT
_ION_EVENT_CONTAINER_END = IonEvent(IonEventType.CONTAINER_END)
_ION_EVENT_RAW_LST_STRUCT_START = IonEvent(
    IonEventType.CONTAINER_START, IonType.STRUCT, annotations=[_system_token(SID_ION_SYMBOL_TABLE)])
_ION_EVENT_RAW_IMPORTS_LIST_START = IonEvent(
    IonEventType.CONTAINER_START, IonType.LIST, field_name=_system_token(SID_IMPORTS))
_ION_EVENT_RAW_SYMBOLS_LIST_START = IonEvent(
    IonEventType.CONTAINER_START, IonType.LIST, field_name=_system_token(SID_SYMBOLS))


@coroutine
def _symbol_table_coroutine(writer_buffer, imports):

    def start_lst():
        write(_ION_EVENT_RAW_LST_STRUCT_START)
        if imports:
            write(_ION_EVENT_RAW_IMPORTS_LIST_START)
            for imported in imports:
                # TODO The system table could be allowed as the first import.
                if imported.table_type is not SHARED_TABLE_TYPE:
                    # TODO This should probably fail at creation of the managed writer coroutine,
                    # but that currently requires two imports iterations.
                    raise IonException('Only shared tables may be imported.')
                write(_ION_EVENT_STRUCT_START)
                write(IonEvent(
                    IonEventType.SCALAR, IonType.STRING, imported.name, field_name=_system_token(SID_NAME)))
                write(IonEvent(
                    IonEventType.SCALAR, IonType.INT, imported.version, field_name=_system_token(SID_VERSION)))
                write(IonEvent(
                    IonEventType.SCALAR, IonType.INT, imported.max_id, field_name=_system_token(SID_MAX_ID)))
                write(_ION_EVENT_CONTAINER_END)
            write(_ION_EVENT_CONTAINER_END)
        return _WRITER_EVENT_NEEDS_INPUT_EMPTY

    def write_symbol(symbol):
        if symbol is None:
            raise IonException('Illegal state: local symbol event with None symbol.')
        try:
            key = symbol.sid
            symbol_text = symbol.text
            if symbol_text is not None:
                key = symbol_text
        except AttributeError:
            assert isinstance(symbol, str)
            key = symbol
            symbol_text = symbol
        token = local_symbols.get(key)
        if token is None:
            assert symbol_text is not None
            token = local_symbols.intern(symbol_text)  # This duplicates the 'get' call...
            write(IonEvent(IonEventType.SCALAR, IonType.STRING, token.text))
        return DataEvent(WriteEventType.NEEDS_INPUT, token)

    # TODO support extending the current LST (by importing it)
    # TODO symtab locking?
    local_symbols = None
    symbol_writer = _raw_binary_writer(writer_buffer)
    write = symbol_writer.send
    has_local_symbols = False
    write_result = None
    while True:
        symbol_event, self = (yield write_result)
        if symbol_event.event_type is _SymbolEventType.START_LST:
            write_event = start_lst()
            local_symbols = SymbolTable(LOCAL_TABLE_TYPE, [], imports=imports)
        elif symbol_event.event_type is _SymbolEventType.SYMBOL:
            if local_symbols is None:
                raise IonException('Illegal state: local symbol table not started.')
            if not has_local_symbols:
                write(_ION_EVENT_RAW_SYMBOLS_LIST_START)
                has_local_symbols = True
            write_event = write_symbol(symbol_event.symbol)
        elif symbol_event.event_type is _SymbolEventType.FINISH:
            # If there are no local symbols or imports, there is no need for an explicit LST.
            if has_local_symbols or imports:
                if has_local_symbols:
                    write(_ION_EVENT_CONTAINER_END)  # End the symbols list.
                write(_ION_EVENT_CONTAINER_END)  # End the symbol table struct.
                for partial in _drain(symbol_writer, _ION_EVENT_STREAM_END):
                    yield partial_transition(partial.data, self)
            write_event = NOOP_WRITER_EVENT
        else:
            raise TypeError('Invalid event: %s' % symbol_event)
        write_result = Transition(write_event, self)


@coroutine
def _managed_binary_writer_coroutine(imports):

    def init():
        _value_writer = _raw_binary_writer(BufferTree())
        _symbol_writer = _raw_symbol_writer(BufferTree(), imports)
        return _value_writer, _symbol_writer

    def intern_symbol(sym):
        return symbol_writer.send(_SymbolEvent(_SymbolEventType.SYMBOL, sym)).data

    def intern_symbols(event):
        field_name = event.field_name
        annotations = event.annotations
        if field_name or field_name == '':
            event = event.derive_field_name(intern_symbol(field_name))
        if annotations:
            event = event.derive_annotations(
                [intern_symbol(annotation) for annotation in annotations])
        if event.ion_type is IonType.SYMBOL and event.value is not None:
            event = event.derive_value(intern_symbol(ion_event.value))
        return event

    value_writer, symbol_writer = init()
    write_result = None
    has_written_values = False
    ivm_needed = True
    while True:
        ion_event, self = (yield write_result)
        if ion_event.event_type is IonEventType.VERSION_MARKER:
            if has_written_values:
                # TODO This could be handled by flushing first.
                raise IonException('Unable to write IVM before STREAM_END')
            else:
                if ivm_needed:
                    yield partial_transition(_IVM, self)
                ivm_needed = False
            write_event = _WRITER_EVENT_NEEDS_INPUT_EMPTY
        elif ion_event.event_type is IonEventType.STREAM_END:
            if has_written_values:
                for partial in _drain(symbol_writer, _SYMBOL_EVENT_FINISH):
                    yield partial_transition(partial.data, self)
                for partial in _drain(value_writer, _ION_EVENT_STREAM_END):
                    yield partial_transition(partial.data, self)
                value_writer, symbol_writer = init()
                has_written_values = False
            write_event = NOOP_WRITER_EVENT
            ivm_needed = True
        else:
            # Intern any symbols and delegate to the raw writer.
            if not has_written_values:
                if ivm_needed:
                    yield partial_transition(_IVM, self)
                ivm_needed = False
                symbol_writer.send(_SYMBOL_EVENT_START_LST)
                has_written_values = True
            ion_event = intern_symbols(ion_event)
            write_event = value_writer.send(ion_event)
        write_result = Transition(write_event, self)


def _raw_symbol_writer(writer_buffer, imports):
    """Returns a raw binary symbol table writer co-routine.

    Keyword Args:
        writer_buffer (BufferTree): The buffer in which this writer's values will be stored.
        imports (Optional[Sequence[SymbolTable]]): A list of shared symbol tables to
            be used by this writer.

    Yields:
        DataEvent: serialization events to write out

        Receives :class:`amazon.ion.core.IonEvent`.
    """
    return writer_trampoline(_symbol_table_coroutine(writer_buffer, imports))


def binary_writer(imports=None):
    """Returns a binary writer co-routine.

    Keyword Args:
        imports (Optional[Sequence[SymbolTable]]): A list of shared symbol tables
            to be used by this writer.

    Yields:
        DataEvent: serialization events to write out

        Receives :class:`amazon.ion.core.IonEvent`.
    """
    return writer_trampoline(_managed_binary_writer_coroutine(imports))




# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import platform
import shutil
import sys
from subprocess import check_call
from os.path import join, abspath, isdir, dirname

_PYPY = hasattr(sys, 'pypy_translation_info')
_OS = platform.system()
_WIN = _OS == 'Windows'
_MAC = _OS == 'Darwin'
_LINUX = _OS == 'Linux'

_C_EXT_DEPENDENCY_DIR = abspath(join(dirname(os.path.abspath(__file__)), 'amazon/ion/ion-c-build'))
_C_EXT_DEPENDENCY_LIB_LOCATION = abspath(join(_C_EXT_DEPENDENCY_DIR, 'lib'))
_C_EXT_DEPENDENCY_INCLUDES_DIR = abspath(join(_C_EXT_DEPENDENCY_DIR, 'include'))
_C_EXT_DEPENDENCY_INCLUDES_LOCATIONS = {
    'ionc': abspath(join(_C_EXT_DEPENDENCY_INCLUDES_DIR, 'ionc')),
    'decNumber': abspath(join(_C_EXT_DEPENDENCY_INCLUDES_DIR, 'decNumber'))
}
_CURRENT_ION_C_DIR = './ion-c'

_IONC_REPO_URL = "https://github.com/amazon-ion/ion-c.git"
_IONC_DIR = abspath(join(dirname(os.path.abspath(__file__)), 'ion-c'))
_IONC_LOCATION = abspath(join(dirname(os.path.abspath(__file__)), 'ion-c', 'build', 'release'))
_IONC_INCLUDES_LOCATIONS = {
    'ionc': abspath(join(dirname(os.path.abspath(__file__)), 'ion-c', 'ionc', 'include', 'ionc')),
    'decNumber': abspath(join(dirname(os.path.abspath(__file__)), 'ion-c', 'decNumber', 'include', 'decNumber'))
}

_LIB_PREFIX = 'lib'

_LIB_SUFFIX_MAC = '.dylib'
_LIB_SUFFIX_WIN = '.lib'
_LIB_SUFFIX_LINUX = '.so'


def _get_lib_name(name):
    if _MAC:
        return '%s%s%s' % (_LIB_PREFIX, name, _LIB_SUFFIX_MAC)
    elif _LINUX:
        return '%s%s%s' % (_LIB_PREFIX, name, _LIB_SUFFIX_LINUX)
    elif _WIN:
        return '%s%s' % (name, _LIB_SUFFIX_WIN)


def _library_exists():
    return _library_exists_helper('ionc') and _library_exists_helper('decNumber')


def _library_exists_helper(name):
    return os.path.exists(join(_IONC_INCLUDES_LOCATIONS[name])) \
        and os.path.exists(join(_IONC_LOCATION, name))


def readonly_handler(func, path, execinfo):
    os.chmod(path, 128)
    func(path)


def _download_ionc():
    try:
        # Install ion-c.
        if isdir(_CURRENT_ION_C_DIR):
            shutil.rmtree(_CURRENT_ION_C_DIR, onerror=readonly_handler)

        # TODO submodule does not work, need to be fixed.
        # check_call(['git', 'submodule', 'update', '--init', '--recursive'])
        check_call(['git', 'clone', '--recurse-submodules', _IONC_REPO_URL, 'ion-c'])

        os.chdir(_CURRENT_ION_C_DIR)

        # Initialize submodule.
        check_call(['git', 'submodule', 'update', '--init'])

        # Build ion-c.
        _build_ionc()

        os.chdir('../')
        return True
    except:
        if isdir(_IONC_DIR):
            shutil.rmtree(_IONC_DIR, onerror=readonly_handler)
        print('ionc build error: Unable to build ion-c library.')
        return False


def _build_ionc():
    if _WIN:
        _build_ionc_win()
    elif _MAC or _LINUX:
        _build_ionc_mac_and_linux()


def _move_ionc():
    # move ion-c to output dir.
    if _WIN:
        _move_lib_win('ionc')
        _move_lib_win('decNumber')
    elif _MAC or _LINUX:
        _move_lib_mac_and_linux('ionc')
        _move_lib_mac_and_linux('decNumber')


def _build_ionc_win():
    # check_call('cmake -G \"Visual Studio 15 2017 Win64\"')
    # check_call('cmake -G \"Visual Studio 16 2019\"')
    check_call('cmake -G \"Visual Studio 17 2022\"')
    check_call('cmake --build . --config Release')


def _move_lib_win(name):
    """
    Move library and its include files to ion-c-build/lib and ion-c-build/include respectively.
    """
    for f in os.listdir(_IONC_INCLUDES_LOCATIONS[name]):
        shutil.copy(join(_IONC_INCLUDES_LOCATIONS[name], f), _C_EXT_DEPENDENCY_INCLUDES_LOCATIONS[name])

    lib_path = join(_IONC_DIR, name, 'Release', '%s%s' % (name, _LIB_SUFFIX_WIN))
    shutil.copy(lib_path, _C_EXT_DEPENDENCY_LIB_LOCATION)


def _build_ionc_mac_and_linux():
    # build ion-c.
    check_call(['./build-release.sh'])


def _move_lib_mac_and_linux(name):
    """
    Move library and its include files to ion-c-build/lib and ion-c-build/include respectively.
    """
    for f in os.listdir(_IONC_INCLUDES_LOCATIONS[name]):
        shutil.copy(join(_IONC_INCLUDES_LOCATIONS[name], f), _C_EXT_DEPENDENCY_INCLUDES_LOCATIONS[name])

    dir_path = join(_IONC_LOCATION, name)
    for file in os.listdir(dir_path):
        file_path = join(dir_path, file)
        if _LINUX:
            if file.startswith('%s%s%s' % (_LIB_PREFIX, name, _LIB_SUFFIX_LINUX)):
                shutil.copy(file_path, _C_EXT_DEPENDENCY_LIB_LOCATION)
        elif _MAC:
            if file.endswith(_LIB_SUFFIX_MAC):
                shutil.copy(file_path, _C_EXT_DEPENDENCY_LIB_LOCATION)


def move_build_lib_for_distribution():
    # Create a directory to store build output.
    if isdir(_C_EXT_DEPENDENCY_DIR):
        shutil.rmtree(_C_EXT_DEPENDENCY_DIR, onerror=readonly_handler)
    os.mkdir(_C_EXT_DEPENDENCY_DIR)
    os.mkdir(_C_EXT_DEPENDENCY_LIB_LOCATION)
    os.mkdir(_C_EXT_DEPENDENCY_INCLUDES_DIR)
    os.mkdir(_C_EXT_DEPENDENCY_INCLUDES_LOCATIONS['ionc'])
    os.mkdir(_C_EXT_DEPENDENCY_INCLUDES_LOCATIONS['decNumber'])
    # Move ion-c binaries to ion-c-build
    _move_ionc()


def _check_dependencies():
    try:
        check_call(['git', '--version'])
        check_call(['cmake', '--version'])
        # TODO add more dependency check here.
    except:
        print('ion-c build error: Missing dependencies.')
        return False
    return True


def _install_ionc():
    if _PYPY:  # This is pointless if running with PyPy, which doesn't support CPython extensions anyway.
        return False

    if not _check_dependencies():
        return False

    if not _library_exists():
        if not _download_ionc():
            return False
    move_build_lib_for_distribution()

    return True


if __name__ == '__main__':
    _install_ionc()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

import itertools
from typing import NamedTuple, Any

import tests

from amazon.ion.core import IonType


class _P(NamedTuple):
    type: str
    expected: Any

    def __str__(self):
        return '{name} - {expected}'.format(name=self.type.name, expected=str(self.expected).upper())


def in_and_out_params(in_types, out_types):
    return tuple(itertools.chain(
        (_P(t, False) for t in out_types),
        (_P(t, True) for t in in_types)
    ))

_IN_TYPES = {IonType.SYMBOL, IonType.STRING}
_OUT_TYPES = set(IonType) - _IN_TYPES


@tests.parametrize(*in_and_out_params(_IN_TYPES, _OUT_TYPES))
def test_is_text(p):
    assert p.expected == p.type.is_text


_IN_TYPES = {IonType.CLOB, IonType.BLOB}
_OUT_TYPES = set(IonType) - _IN_TYPES


@tests.parametrize(*in_and_out_params(_IN_TYPES, _OUT_TYPES))
def test_is_lob(p):
    assert p.expected == p.type.is_lob


_IN_TYPES = {IonType.LIST, IonType.SEXP, IonType.STRUCT}
_OUT_TYPES = set(IonType) - _IN_TYPES


@tests.parametrize(*in_and_out_params(_IN_TYPES, _OUT_TYPES))
def test_is_container(p):
    assert p.expected == p.type.is_container



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from datetime import datetime
from decimal import Decimal
from typing import Type, NamedTuple

from tests import parametrize, listify
from tests.event_aliases import *

from amazon.ion.core import Timestamp, TimestampPrecision
from amazon.ion.symbols import SymbolToken
from amazon.ion.simple_types import is_null, IonPyNull, IonPyBool, IonPyInt, IonPyFloat, \
    IonPyDecimal, IonPyTimestamp, IonPyText, IonPyBytes, \
    IonPyList, IonPyDict, IonPySymbol
from amazon.ion.equivalence import ion_equals
from amazon.ion.simpleion import _ion_type, _FROM_TYPE

_TEST_FIELD_NAME = SymbolToken('foo', 10)
_TEST_ANNOTATIONS = (SymbolToken('bar', 11),)


class _P(NamedTuple):
    desc: str
    type: Type[object]
    event: IonEvent

    def __str__(self):
        return self.desc


_EVENT_TYPES = [
    (IonPyNull, e_null()),
    (IonPyNull, e_int(None)),
    (IonPyBool, e_bool(True)),
    (IonPyInt, e_int(65000)),
    (IonPyInt, e_int(2 ** 64 + 1)),
    (IonPyFloat, e_float(1e0)),
    (IonPyDecimal, e_decimal(Decimal('1.1'))),
    (IonPyTimestamp, e_timestamp(datetime(2012, 1, 1))),
    (IonPyTimestamp, e_timestamp(Timestamp(2012, 1, 1, precision=TimestampPrecision.DAY))),
    (IonPySymbol, e_symbol(SymbolToken(u'Hola', None, None))),
    (IonPyText, e_string(u'Hello')),
    (IonPyBytes, e_clob(b'Goodbye')),

    # Technically this is not how we shape the event, but the container types don't care
    (IonPyList, e_start_sexp([1, 2, 3])),
    (IonPyDict, e_start_struct({u'hello': u'world'}))
]


@listify
def event_type_parameters():
    for cls, event in _EVENT_TYPES:
        yield _P(
            desc='EVENT TO %s - %s - %r' % (cls.__name__, event.ion_type.name, event.value),
            type=cls,
            event=event.derive_annotations(_TEST_ANNOTATIONS).derive_field_name(_TEST_FIELD_NAME)
        )


@parametrize(*event_type_parameters())
def test_event_types(p):
    value = p.event.value
    ion_type = p.event.ion_type

    event_output = p.type.from_event(p.event)
    value_output = p.type.from_value(ion_type, value, p.event.annotations)
    to_event_output = value_output.to_event(p.event.event_type, p.event.field_name, in_struct=True, depth=p.event.depth)
    if p.event.ion_type.is_container:
        # compensate for abuse of IonEvent.value, which is intended to be None for CONTAINER_START events,
        # but is populated and relied upon by the logic of this test code
        assert to_event_output.value is None
        to_event_output = to_event_output._replace(value=p.event.value)
    assert p.event == to_event_output

    if p.type is IonPyNull:
        # Null is a special case due to the non-extension of NoneType.
        assert not event_output
        assert not value_output
        assert is_null(event_output)
        assert is_null(value_output)
        assert ion_equals(event_output, value_output)
    else:
        # Make sure we don't change value equality symmetry.
        assert event_output == value
        assert value == event_output

        assert value_output == value
        assert value == value_output

        # Derive a new event from just the value because equality is stricter in some cases.
        value_event = e_scalar(ion_type, value)
        output_event = e_scalar(ion_type, event_output)
        assert value_event == output_event

    assert event_output.ion_type is ion_type
    assert p.event.annotations == event_output.ion_annotations

    assert value_output.ion_type is ion_type
    assert p.event.annotations == value_output.ion_annotations


def test_subclass_types():
    class Foo(dict):
        pass

    assert _ion_type(Foo(), _FROM_TYPE) is IonType.STRUCT



from os.path import abspath, join, dirname

from amazon.ionbenchmark.benchmark_spec import BenchmarkSpec


def _generate_test_path(p):
    return join(dirname(abspath(__file__)), 'benchmark_sample_data', p)


_minimal_params = {'format': "ion_text", 'input_file': "cat.ion"}
_minimal_spec = BenchmarkSpec(_minimal_params, working_directory=_generate_test_path("sample_spec"))


def test_get_input_file_size():
    assert _minimal_spec.get_input_file_size() == 161


def test_get_format():
    assert _minimal_spec.get_format() == 'ion_text'


def test_get_command():
    assert _minimal_spec.get_command() == 'read'


def test_get_api():
    assert _minimal_spec.get_api() == 'load_dump'


def test_get_name():
    assert _minimal_spec.get_name() == '(ion_text,loads,cat.ion)'


def test_defaults_and_overrides_applied_in_correct_order():
    a = {'c': 1, 'b': 10, 'a': 100}
    b = {'c': 2, 'b': 20}
    c = {'c': 3, **_minimal_params}

    spec = BenchmarkSpec(user_defaults=a, params=b, user_overrides=c)

    # From tool default
    assert spec['api'] == 'load_dump'
    # From user default
    assert spec['a'] == 100
    # From params
    assert spec['b'] == 20
    # From user override
    assert spec['c'] == 3



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
from typing import NamedTuple, Any

from pytest import raises

from tests import parametrize, is_exception, noop_manager

from amazon.ion.util import unicode_iter


def _unichr_list(val):
    return list(ord(x) for x in val)


class _P(NamedTuple):
    desc: str
    input: str
    expected: Any

    def __str__(self):
        return self.desc


@parametrize(
    _P('ASCII', u'abcd', _unichr_list(u'abcd')),
    _P('BMP SEQUENCE', u'abcd\u3000', _unichr_list(u'abcd\u3000')),
    _P('NON-BMP SEQUENCE', u'abcd\U0001f4a9', _unichr_list(u'abcd') + [0x1F4A9]),
    _P('UNPAIRED LOW', u'\udc00', ValueError),
    _P('UNPAIRED HIGH AT END', u'\ud800', ValueError),
    _P('UNPAIRED HIGH IN MID', u'\ud800a', ValueError),
)
def test_unicode_iter(p):
    ctx = noop_manager()
    if is_exception(p.expected):
        ctx = raises(p.expected)

    with ctx:
        actual = list(unicode_iter(p.input))

        if not is_exception(p.expected):
            assert p.expected == actual
